import ctypes
import functools
import inspect
import warnings
import decimal as _decimal
from typing import Optional, Type, Union
from datetime import date as _date
from datetime import datetime as _datetime
from datetime import timedelta as _timedelta

import pandas as pd
import numpy as np
from pandas.tseries import offsets
from packaging.version import parse as parse_version

import onetick.py as otp
from onetick.py.otq import otq, pyomd
from onetick.py.compatibility import has_timezone_parameter
from onetick.py.core._internal._op_utils.every_operand import every_operand
from onetick.py.utils import get_tzfile_by_name, get_timezone_from_datetime
from onetick.py.docs.utils import is_windows

# --------------------------------------------------------------- #
# TYPES IMPLEMENTATION
# --------------------------------------------------------------- #


class OTPBaseTimeStamp(type):
    pass


class _nsectime(OTPBaseTimeStamp):
    def __str__(cls):
        return "nsectime"


class nsectime(int, metaclass=_nsectime):
    """
    OneTick data type representing datetime with nanoseconds precision.
    Can be used to specify otp.Source column type when converting columns or creating new ones.
    Note that this constructor creates datetime value in GMT timezone
    and doesn't take into account the timezone with which the query is executed.

    Examples
    --------
    >>> t = otp.Tick(A=0)
    >>> t['A'] = t['A'].apply(otp.nsectime)
    >>> t['B'] = otp.nsectime(24 * 60 * 60 * 1000 * 1000 * 1000 + 2)
    >>> t.schema
    {'A': <class 'onetick.py.types.nsectime'>, 'B': <class 'onetick.py.types.nsectime'>}
    >>> otp.run(t)
            Time                   A                             B
    0 2003-12-01 1969-12-31 19:00:00 1970-01-01 19:00:00.000000002
    """
    def __str__(self):
        return super().__repr__()

    def __repr__(self):
        return f'{self.__class__.__name__}({self})'


class _msectime(OTPBaseTimeStamp):
    def __str__(cls):
        return "msectime"


class msectime(int, metaclass=_msectime):
    """
    OneTick data type representing datetime with milliseconds precision.
    Can be used to specify otp.Source column type when converting columns or creating new ones.
    Note that this constructor creates datetime value in GMT timezone
    and doesn't take into account the timezone with which the query is executed.

    Examples
    --------
    >>> t = otp.Tick(A=1)
    >>> t = t.table(A=otp.msectime)
    >>> t['B'] = otp.msectime(2)
    >>> t.schema
    {'A': <class 'onetick.py.types.msectime'>, 'B': <class 'onetick.py.types.msectime'>}
    >>> otp.run(t)
            Time                       A                       B
    0 2003-12-01 1969-12-31 19:00:00.001 1969-12-31 19:00:00.002
    """
    def __str__(self):
        return super().__repr__()

    def __repr__(self):
        return f'{self.__class__.__name__}({self})'


class OTPBaseTimeOffset:
    datepart = "'invalid'"  # that is just base class for other dateparts
    n = 0
    delta = pd.Timedelta(seconds=0)

    def get_offset(self):
        return self.n, self.datepart[1:-1]


class ExpressionDefinedTimeOffset(OTPBaseTimeOffset):
    def __init__(self, datepart, n):
        self.datepart = datepart
        self.n = n

        from onetick.py.core.column_operations.base import _Operation

        def proxy_wrap(attr):
            def f(self, *args, **kwargs):
                return getattr(self.n, attr)(*args, **kwargs)
            return f

        for attr, value in inspect.getmembers(_Operation, callable):
            if attr in {'__class__', '__init__', '__new__', '__init_subclass__', '__dir__',
                        '__getattribute__', '__getattr__', '__delattr__', '__setattr__',
                        '__subclasshook__', '__sizeof__', '__str__', '__repr__'}:
                continue
            setattr(ExpressionDefinedTimeOffset, attr, proxy_wrap(attr))


# ---------------------------- #
# Implement datepart units

def _construct_dpf(dp_class, str_repr=None, **dp_class_params):
    """ construct a datepart factory """

    if str_repr is None:
        str_repr = dp_class.__name__.lower()

    class _DatePartCls(dp_class, OTPBaseTimeOffset):
        datepart = f"'{str_repr}'"

    def _factory(n):
        from onetick.py.core.column_operations._methods.methods import is_arithmetical
        from onetick.py.core.column import _Column

        if isinstance(n, int):
            if dp_class_params:
                return _DatePartCls(**dp_class_params) * n
            return _DatePartCls(n)
        if is_arithmetical(n):
            n = _process_datediff(n)
            return ExpressionDefinedTimeOffset(_DatePartCls.datepart, n)
        if isinstance(n, _Column):
            return ExpressionDefinedTimeOffset(_DatePartCls.datepart, n)
        raise ValueError("Unknown type was passed as arg, integer constant or column or expression is expected here")

    def _process_datediff(n):

        n_time_operand = _get_n_time_operand(n)
        if n_time_operand:
            # check if otp.Hour(date1 - date2) is called, return a number of hours between two days in such ways
            from onetick.py.core.column_operations._methods.methods import sub, _wrap_object
            from onetick.py.core.column_operations.base import _Operation
            from onetick.py.core.column import _LagOperator

            available_types = (_Operation, _LagOperator)
            if (getattr(n, "_op_func", sub) and len(n._op_params) == 2
                    and isinstance(n._op_params[0], available_types) and isinstance(n._op_params[1], available_types)):
                def _datediff(*args):
                    args = ', '.join(map(_wrap_object, args))
                    return f'DATEDIFF({_DatePartCls.datepart}, {args}, _TIMEZONE)', int
                return _Operation(_datediff, [n._op_params[1], n._op_params[0]])
            else:
                raise ValueError(
                    "Date arithmetic operations (except date2-date1, which calculate an amount of "
                    "periods between two dates) are not accepted in TimeOffset constructor"
                )
        return n

    def _get_n_time_operand(n):
        from onetick.py.core.column_operations._methods.op_types import are_time

        result = 0
        for op in every_operand(n):
            if are_time(get_object_type(op)):
                result += 1
        return result

    return _factory


def _construct_float_dpf(dp_class, dp_float_class, power, str_repr=None, float_str_repr=None, **dp_class_params):
    def _factory(n):
        if isinstance(n, float):
            return _construct_dpf(dp_float_class, float_str_repr, **dp_class_params)(int(n * (10 ** power)))
        else:
            return _construct_dpf(dp_class, str_repr, **dp_class_params)(n)

    return _factory


_add_examples_to_docs = functools.partial(
    """
    Object representing {0}'s datetime offset.

    Can be added to or subtracted from:

    * :py:class:`otp.datetime <onetick.py.datetime>` objects
    * :py:class:`Source <onetick.py.Source>` columns of datetime type

    Parameters
    ----------
    n: int, :class:`~onetick.py.Column`, :class:`~onetick.py.Operation`{additional_types}
        Offset integer value or column of :class:`~onetick.py.Source`.
        The only :class:`~onetick.py.Operation` supported is
        subtracting one datetime column from another. See example below.{additional_notes}

    Examples
    --------
    {1}
    """.format, additional_types="", additional_notes="",
)

_float_dpf_types = ", float"
_float_dpf_notes = """
        Offset could be ``float`` to pass a fractional time unit value."""


Year = _construct_dpf(offsets.DateOffset, "year", years=1)
Year.__doc__ = _add_examples_to_docs('year', """
    Add to or subtract from :py:class:`otp.datetime <onetick.py.datetime>` object:

    >>> otp.datetime(2012, 12, 12, 12) + otp.Year(1)
    2013-12-12 12:00:00
    >>> otp.datetime(2012, 12, 12, 12) - otp.Year(1)
    2011-12-12 12:00:00

    Use offset in columns:

    >>> t = otp.Tick(A=1)
    >>> t['T'] = otp.datetime(2012, 12, 12, 12)
    >>> t['T'] += otp.Year(t['A'])
    >>> otp.run(t)
            Time                   T  A
    0 2003-12-01 2013-12-12 12:00:00  1

    Use it to calculate difference between two dates:

    >>> t = otp.Tick(A=otp.dt(2022, 1, 1), B=otp.dt(2023, 1, 1))
    >>> t['DIFF'] = otp.Year(t['B'] - t['A'])
    >>> otp.run(t)
            Time           A           B  DIFF
    0 2003-12-01  2022-01-01  2023-01-01     1
""")

Quarter = _construct_dpf(offsets.DateOffset, "quarter", months=3)
Quarter.__doc__ = _add_examples_to_docs('quarter', """
    Add to or subtract from :py:class:`otp.datetime <onetick.py.datetime>` object:

    >>> otp.datetime(2012, 12, 12, 12) + otp.Quarter(1)
    2013-03-12 12:00:00
    >>> otp.datetime(2012, 12, 12, 12) - otp.Quarter(1)
    2012-09-12 12:00:00

    Use offset in columns:

    >>> t = otp.Tick(A=1)
    >>> t['T'] = otp.datetime(2012, 12, 12, 12, tz='GMT')
    >>> t['T'] += otp.Quarter(t['A'])
    >>> otp.run(t, start=otp.datetime(2003, 12, 2), end=otp.datetime(2003, 12, 3), timezone='GMT')
            Time                   T  A
    0 2003-12-02 2013-03-12 12:00:00  1

    Use it to calculate difference between two dates:

    >>> t = otp.Tick(A=otp.dt(2022, 1, 1), B=otp.dt(2023, 1, 1))
    >>> t['DIFF'] = otp.Quarter(t['B'] - t['A'])
    >>> otp.run(t)
            Time           A           B  DIFF
    0 2003-12-01  2022-01-01  2023-01-01     4
""")

Month = _construct_dpf(offsets.DateOffset, "month", months=1)
Month.__doc__ = _add_examples_to_docs('month', """
    Add to or subtract from :py:class:`otp.datetime <onetick.py.datetime>` object:

    >>> otp.datetime(2012, 12, 12, 12) + otp.Month(1)
    2013-01-12 12:00:00
    >>> otp.datetime(2012, 12, 12, 12) - otp.Month(1)
    2012-11-12 12:00:00

    Use offset in columns:

    >>> t = otp.Tick(A=1)
    >>> t['T'] = otp.datetime(2012, 12, 12, 12)
    >>> t['T'] += otp.Month(t['A'])
    >>> otp.run(t)
            Time                   T  A
    0 2003-12-01 2013-01-12 12:00:00  1

    Use it to calculate difference between two dates:

    >>> t = otp.Tick(A=otp.dt(2022, 1, 1), B=otp.dt(2023, 1, 1))
    >>> t['DIFF'] = otp.Month(t['B'] - t['A'])
    >>> otp.run(t)
            Time           A           B  DIFF
    0 2003-12-01  2022-01-01  2023-01-01    12
""")

Week = _construct_dpf(offsets.Week)
Week.__doc__ = _add_examples_to_docs('week', """
    Add to or subtract from :py:class:`otp.datetime <onetick.py.datetime>` object:

    >>> otp.datetime(2012, 12, 12, 12) + otp.Week(1)
    2012-12-19 12:00:00
    >>> otp.datetime(2012, 12, 12, 12) - otp.Week(1)
    2012-12-05 12:00:00

    Use offset in columns:

    >>> t = otp.Tick(A=1)
    >>> t['T'] = otp.datetime(2012, 12, 12, 12)
    >>> t['T'] += otp.Week(t['A'])
    >>> otp.run(t)
            Time                   T  A
    0 2003-12-01 2012-12-19 12:00:00  1

    Use it to calculate difference between two dates:

    >>> t = otp.Tick(A=otp.dt(2022, 1, 1), B=otp.dt(2023, 1, 1))
    >>> t['DIFF'] = otp.Week(t['B'] - t['A'])
    >>> otp.run(t)
            Time           A           B  DIFF
    0 2003-12-01  2022-01-01  2023-01-01    53
""")

Day = _construct_dpf(offsets.Day)
Day.__doc__ = _add_examples_to_docs('day', """
    Add to or subtract from :py:class:`otp.datetime <onetick.py.datetime>` object:

    >>> otp.datetime(2012, 12, 12, 12) + otp.Day(1)
    2012-12-13 12:00:00
    >>> otp.datetime(2012, 12, 12, 12) - otp.Day(1)
    2012-12-11 12:00:00

    Use offset in columns:

    >>> t = otp.Tick(A=1)
    >>> t['T'] = otp.datetime(2012, 12, 12, 12)
    >>> t['T'] += otp.Day(t['A'])
    >>> otp.run(t)
            Time                   T  A
    0 2003-12-01 2012-12-13 12:00:00  1

    Use it to calculate difference between two dates:

    >>> t = otp.Tick(A=otp.dt(2022, 1, 1), B=otp.dt(2023, 1, 1))
    >>> t['DIFF'] = otp.Day(t['B'] - t['A'])
    >>> otp.run(t)
            Time           A           B  DIFF
    0 2003-12-01  2022-01-01  2023-01-01   365
""")

Hour = _construct_dpf(offsets.Hour)
Hour.__doc__ = _add_examples_to_docs('hour', """
    Add to or subtract from :py:class:`otp.datetime <onetick.py.datetime>` object:

    >>> otp.datetime(2012, 12, 12, 12) + otp.Hour(1)
    2012-12-12 13:00:00
    >>> otp.datetime(2012, 12, 12, 12) - otp.Hour(1)
    2012-12-12 11:00:00

    Use offset in columns:

    >>> t = otp.Tick(A=1)
    >>> t['T'] = otp.datetime(2012, 12, 12, 12)
    >>> t['T'] += otp.Hour(t['A'])
    >>> otp.run(t)
            Time                   T  A
    0 2003-12-01 2012-12-12 13:00:00  1

    Use it to calculate difference between two dates:

    >>> t = otp.Tick(A=otp.dt(2022, 1, 1), B=otp.dt(2022, 1, 2))
    >>> t['DIFF'] = otp.Hour(t['B'] - t['A'])
    >>> otp.run(t)
            Time           A           B  DIFF
    0 2003-12-01  2022-01-01  2022-01-02    24
""")

Minute = _construct_dpf(offsets.Minute)
Minute.__doc__ = _add_examples_to_docs('minute', """
    Add to or subtract from :py:class:`otp.datetime <onetick.py.datetime>` object:

    >>> otp.datetime(2012, 12, 12, 12) + otp.Minute(1)
    2012-12-12 12:01:00
    >>> otp.datetime(2012, 12, 12, 12) - otp.Minute(1)
    2012-12-12 11:59:00

    Use offset in columns:

    >>> t = otp.Tick(A=1)
    >>> t['T'] = otp.datetime(2012, 12, 12, 12)
    >>> t['T'] += otp.Minute(t['A'])
    >>> otp.run(t)
            Time                   T  A
    0 2003-12-01 2012-12-12 12:01:00  1

    Use it to calculate difference between two dates:

    >>> t = otp.Tick(A=otp.dt(2022, 1, 1), B=otp.dt(2022, 1, 1, 1))
    >>> t['DIFF'] = otp.Minute(t['B'] - t['A'])
    >>> otp.run(t)
            Time           A                    B  DIFF
    0 2003-12-01  2022-01-01  2022-01-01 01:00:00    60
""")

Second = _construct_float_dpf(offsets.Second, offsets.Nano, 9, None, "nanosecond")
Second.__doc__ = _add_examples_to_docs('second', """
    Add to or subtract from :py:class:`otp.datetime <onetick.py.datetime>` object:

    >>> otp.datetime(2012, 12, 12, 12) + otp.Second(1)
    2012-12-12 12:00:01
    >>> otp.datetime(2012, 12, 12, 12) - otp.Second(1)
    2012-12-12 11:59:59

    Using ``float`` value to pass nanoseconds:

    >>> otp.datetime(2012, 12, 12, 12) + otp.Second(1.000000123)
    2012-12-12 12:00:01.000000123

    Use offset in columns:

    >>> t = otp.Tick(A=1)
    >>> t['T'] = otp.datetime(2012, 12, 12, 12)
    >>> t['T'] += otp.Second(t['A'])
    >>> otp.run(t)
            Time                   T  A
    0 2003-12-01 2012-12-12 12:00:01  1

    Use it to calculate difference between two dates:

    >>> t = otp.Tick(A=otp.dt(2022, 1, 1), B=otp.dt(2022, 1, 1, 0, 1))
    >>> t['DIFF'] = otp.Second(t['B'] - t['A'])
    >>> otp.run(t)
            Time           A                    B  DIFF
    0 2003-12-01  2022-01-01  2022-01-01 00:01:00    60
""", additional_types=_float_dpf_types, additional_notes=_float_dpf_notes)

Milli = _construct_float_dpf(offsets.Milli, offsets.Nano, 6, "millisecond", "nanosecond")
Milli.__doc__ = _add_examples_to_docs('millisecond', """
    Add to or subtract from :py:class:`otp.datetime <onetick.py.datetime>` object:

    >>> otp.datetime(2012, 12, 12, 12) + otp.Milli(1)
    2012-12-12 12:00:00.001000
    >>> otp.datetime(2012, 12, 12, 12) - otp.Milli(1)
    2012-12-12 11:59:59.999000

    Using ``float`` value to pass nanoseconds:

    >>> otp.datetime(2012, 12, 12, 12) + otp.Milli(1.000123)
    2012-12-12 12:00:00.001000123

    Use offset in columns:

    >>> t = otp.Tick(A=1)
    >>> t['T'] = otp.datetime(2012, 12, 12, 12)
    >>> t['T'] += otp.Milli(t['A'])
    >>> otp.run(t)
            Time                       T  A
    0 2003-12-01 2012-12-12 12:00:00.001  1

    Use it to calculate difference between two dates:

    >>> t = otp.Tick(A=otp.dt(2022, 1, 1), B=otp.dt(2022, 1, 1, 0, 0, 1))
    >>> t['DIFF'] = otp.Milli(t['B'] - t['A'])
    >>> otp.run(t)
            Time           A                    B  DIFF
    0 2003-12-01  2022-01-01  2022-01-01 00:00:01  1000
""", additional_types=_float_dpf_types, additional_notes=_float_dpf_notes)

# microseconds are not supported yet

Nano = _construct_dpf(offsets.Nano, "nanosecond")
Nano.__doc__ = _add_examples_to_docs('nanosecond', """
    Add to or subtract from :py:class:`otp.datetime <onetick.py.datetime>` object:

    >>> otp.datetime(2012, 12, 12, 12) + otp.Nano(1)
    2012-12-12 12:00:00.000000001
    >>> otp.datetime(2012, 12, 12, 12) - otp.Nano(1)
    2012-12-12 11:59:59.999999999

    Use offset in columns:

    >>> t = otp.Tick(A=1)
    >>> t['T'] = otp.datetime(2012, 12, 12, 12)
    >>> t['T'] += otp.Nano(t['A'])
    >>> otp.run(t)
            Time                             T  A
    0 2003-12-01 2012-12-12 12:00:00.000000001  1

    Use it to calculate difference between two dates:

    >>> t = otp.Tick(A=otp.dt(2022, 1, 1), B=otp.dt(2022, 1, 1, 0, 0, 1))
    >>> t['DIFF'] = otp.Nano(t['B'] - t['A'])
    >>> otp.run(t)
            Time           A                    B        DIFF
    0 2003-12-01  2022-01-01  2022-01-01 00:00:01  1000000000
""")

# ---------------------------- #


class _inner_string(type):
    def __str__(cls):
        if cls.length is Ellipsis:
            return "varstring"
        if cls.length:
            return f"string[{cls.length}]"
        else:
            return "string"

    def __repr__(cls):
        return str(cls)

    # We have ot use functools.cache, because 'class' in python is an object,
    # and _inner_str for the same item is different for every call,
    # but we want to make str[1024] be equal to another str[1024]
    @functools.lru_cache(maxsize=None)  # noqa: W1518
    def __getitem__(cls, item):

        if (not isinstance(item, int) or item < 1) and item is not Ellipsis:
            raise TypeError("It is not allowed to have non numeric index")

        class _inner_str(string):  # type: ignore[misc]
            length = item

            def __len__(self):
                return self.__class__.length

        return _inner_str


class string(str, metaclass=_inner_string):  # type: ignore[misc]
    """
    OneTick data type representing string with length and varstring.
    To set string length use ``__getitem__``.
    If the length is not set then the :py:attr:`~DEFAULT_LENGTH` length is used by default.
    In this case using ``otp.string`` is the same as using ``str``.
    If the length is set to Ellipse it represents varstring. Varstring is used for returning variably sized strings.

    Note
    ----
    If you try to set value with length x to string[y] and x > y, value will be truncated to y length.

    Attributes
    ----------
    DEFAULT_LENGTH: int
        default length of the string when the length is not specified

    Examples
    --------

    Adding new fields with :class:`otp.string <string>` type:

    >>> t = otp.Tick(A=otp.string[32]('HELLO'))
    >>> t['B'] = otp.string[128]('WORLD')
    >>> t.schema
    {'A': string[32], 'B': string[128]}
    >>> otp.run(t)
            Time      A      B
    0 2003-12-01  HELLO  WORLD

    Note that class :class:`otp.string <string>` is a child of python's ``str`` class,
    so every object passed to constructor is converted to string.
    So it may not work as expected in some cases, e.g. when passing :class:`~onetick.py.Operation` objects
    (because their string representation is the name of the column or OneTick expression).
    In this case it's better to use direct type conversion to get :class:`otp.string <string>` object:

    >>> t = otp.Tick(X=otp.string[256](t['_SYMBOL_NAME']))
    >>> t['Y'] = t['_SYMBOL_NAME'].astype(otp.string[256])
    >>> t.schema
    {'X': string[256], 'Y': string[256]}
    >>> otp.run(t, symbols='AAAA')
            Time             X     Y
    0 2003-12-01  _SYMBOL_NAME  AAAA

    Setting the type of the existing field:

    >>> # OTdirective: skip-snippet:;
    >>> t = otp.Tick(A='a')
    >>> t = t.table(A=otp.string[10])
    >>> t.schema
    {'A': string[10]}

    Example of truncation column value to set string length.

    >>> # OTdirective: skip-snippet:;
    >>> t['A'] *= 100
    >>> t['B'] = t['A'].str.len()
    >>> otp.run(t)
            Time           A   B
    0 2003-12-01  aaaaaaaaaa  10

    Example of string with default length.

    >>> t = otp.Tick(A='a')
    >>> t['A'] *= 100
    >>> t['B'] = t['A'].str.len()
    >>> otp.run(t)
            Time                                                                 A   B
    0 2003-12-01  aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa  64

    Setting Ellipsis as length represents varstring.

    >>> t = otp.Tick(A='a')
    >>> t = t.table(A=otp.string[...])
    >>> t.schema
    {'A': varstring}

    Varstring length is multiplied.

    >>> t['A'] *= 65
    >>> t['B'] = t['A'].str.len()
    >>> otp.run(t)
            Time                                                                  A   B
    0 2003-12-01  aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa  65

    `otp.varstring` is a shortcut:

    >>> t = otp.Tick(A='a')
    >>> t = t.table(A=otp.varstring)
    >>> t.schema
    {'A': varstring}
    """

    DEFAULT_LENGTH = 64
    length = None

    def __repr__(self):
        return f'{self.__class__}({super().__repr__()})'


varstring = string[...]  # type: ignore[type-arg,misc]


class _nan_base(type):
    def __str__(cls):
        return "double"


class _nan(float, metaclass=_nan_base):
    """
    Object that represents NaN (not a number) float value.
    Can be used anywhere where float value is expected.

    Examples
    --------
    >>> t = otp.Ticks({'A': [1.1, 2.2, otp.nan]})
    >>> t['B'] = otp.nan
    >>> t['C'] = t['A'] / 0
    >>> t['D'] = t['A'] + otp.nan
    >>> otp.run(t)
                         Time    A   B    C   D
    0 2003-12-01 00:00:00.000  1.1 NaN  inf NaN
    1 2003-12-01 00:00:00.001  2.2 NaN  inf NaN
    2 2003-12-01 00:00:00.002  NaN NaN  NaN NaN
    """

    __name__ = 'nan'

    def __str__(self):
        return "NAN()"

    def __repr__(self):
        return 'nan'


nan = _nan()


class _inf(float, metaclass=_nan_base):
    """
    Object that represents infinity value.
    Can be used anywhere where float value is expected.

    Examples
    --------
    >>> t = otp.Ticks({'A': [1.1, 2.2, otp.inf]})
    >>> t['B'] = otp.inf
    >>> t['C'] = t['A'] / 0
    >>> t['D'] = t['A'] - otp.inf
    >>> otp.run(t)
                         Time    A    B    C    D
    0 2003-12-01 00:00:00.000  1.1  inf  inf -inf
    1 2003-12-01 00:00:00.001  2.2  inf  inf -inf
    2 2003-12-01 00:00:00.002  inf  inf  inf  NaN
    """

    __name__ = 'inf'

    def __init__(self):
        self._sign = ""  # empty string or '-' for negative infinity

    def __str__(self):
        return f"{self._sign}INFINITY()"

    def __repr__(self):
        return f'{self._sign}inf'

    def __neg__(self):
        result = _inf()
        result._sign = "" if self._sign else "-"
        return result


inf = _inf()


class decimal:
    """
    Object that represents decimal OneTick value.
    Decimal is 128 bit base 10 floating point number.

    Parameters
    ----------
    value: int, float, str
        The value to initialize decimal from.
        Note that float values may be converted with precision lost.

    Examples
    --------

    :py:class:`~onetick.py.types.decimal` objects can be used in tick generators
    and column operations as any other onetick-py type:

    >>> t = otp.Ticks({'A': [otp.decimal(1), otp.decimal(2)]})
    >>> t['B'] = otp.decimal(1.23456789)
    >>> t['C'] = t['A'] / 0
    >>> t['D'] = t['A'] + otp.nan
    >>> otp.run(t)
                         Time    A         B    C    D
    0 2003-12-01 00:00:00.000  1.0  1.234568  inf  NaN
    1 2003-12-01 00:00:00.001  2.0  1.234568  inf  NaN

    Additionally, any arithmetic operation with :py:class:`~onetick.py.types.decimal` object will return
    an :py:class:`~onetick.py.Operation` object:

    >>> t = otp.Tick(A=1)
    >>> t['X'] = otp.decimal(1) / 0
    >>> otp.run(t)
            Time    A    X
    0 2003-12-01    1  inf

    Note that converting from float (first row) may result in losing precision.
    :py:class:`~onetick.py.types.decimal` objects are created from strings or integers, so they don't lose precision:

    >>> t0 = otp.Tick(A=0.1)
    >>> t1 = otp.Tick(A=otp.decimal(0.01))
    >>> t2 = otp.Tick(A=otp.decimal('0.001'))
    >>> t3 = otp.Tick(A=otp.decimal(1) / otp.decimal(10_000))
    >>> t = otp.merge([t0, t1, t2, t3], enforce_order=True)
    >>> t['STR_A'] = t['A'].decimal.str(34)
    >>> otp.run(t)
            Time       A                                 STR_A
    0 2003-12-01  0.1000  0.1000000000000000055511151231257827
    1 2003-12-01  0.0100  0.0100000000000000000000000000000000
    2 2003-12-01  0.0010  0.0010000000000000000000000000000000
    3 2003-12-01  0.0001  0.0001000000000000000000000000000000

    Note that :py:class:`otp.Ticks <onetick.py.Ticks>` will convert everything from string under the hood,
    so even the float values will not lose precision:

    >>> t = otp.Ticks({'A': [0.1, otp.decimal(0.01), otp.decimal('0.001'), otp.decimal(1e-4)]})
    >>> t['STR_A'] = t['A'].decimal.str(34)
    >>> otp.run(t)
                         Time       A                                 STR_A
    0 2003-12-01 00:00:00.000  0.1000  0.1000000000000000000000000000000000
    1 2003-12-01 00:00:00.001  0.0100  0.0100000000000000000000000000000000
    2 2003-12-01 00:00:00.002  0.0010  0.0010000000000000000000000000000000
    3 2003-12-01 00:00:00.003  0.0001  0.0001000000000000000000000000000000
    """
    def __new__(cls, *args, **kwargs):
        # this method dynamically adds properties and methods
        # from otp.Operation class to this one

        # otp.decimal class doesn't fit well in onetick-py type system,
        # so this class is a mix of both type and Operation logic

        # Basically it works like this:
        #   otp.decimal is a OneTick type
        #   otp.decimal(1) is a decimal type object
        # Doing anything with this object returns an otp.Operation:
        #   otp.decimal(1) / 2

        def proxy_wrap(attr, value):
            if callable(value):
                @functools.wraps(value)
                def f(self, *args, **kwargs):
                    op = self.to_operation()
                    return getattr(op, attr)(*args, **kwargs)
                return f
            else:
                @functools.wraps(value)
                def f(self):
                    op = self.to_operation()
                    return getattr(op, attr)
                return property(f)

        for attr, value in inspect.getmembers(otp.Operation):
            # comparison methods are defined by default for some reason,
            # but we want to get them from otp.Operation
            if not hasattr(cls, attr) or attr in ('__lt__', '__le__', '__eq__', '__ne__', '__gt__', '__ge__'):
                setattr(cls, attr, proxy_wrap(attr, value))

        return super().__new__(cls)

    def __init__(self, value):
        from onetick.py.core.column_operations.base import OnetickParameter
        supported_types = (str, int, float, OnetickParameter)
        if not isinstance(value, supported_types):
            raise TypeError(f"Parameter 'value' must be one of these types: {supported_types}, got {type(value)}")
        self.__value = value

    @classmethod
    def _to_onetick_type_string(cls):
        # called by ott.type2str
        return 'decimal'

    def _to_onetick_string(self):
        # called by ott.value2str
        from onetick.py.core.column_operations.base import OnetickParameter
        if isinstance(self.__value, OnetickParameter):
            value = self.__value
        else:
            value = str(self.__value)
        return f'STRING_TO_DECIMAL({value2str(value)})'

    def to_operation(self):
        return otp.Operation(op_str=self._to_onetick_string(), dtype=decimal)

    def __str__(self):
        # called by otp.CSV, we don't need to convert the value with OneTick functions in this case
        return str(self.__value)

    def __repr__(self):
        return f"{self.__class__.__name__}({value2str(self.__value)})"

    def __format__(self, __format_spec: str) -> str:
        return _decimal.Decimal(self.__value).__format__(__format_spec)

# --------------------------------------------------------------- #
# AUXILIARY FUNCTIONS
# --------------------------------------------------------------- #


def is_type_basic(dtype):
    return dtype in (
        int,
        float,
        str,
        byte,
        short,
        uint,
        ulong,
        _int,
        long,
        nsectime,
        msectime,
        decimal,
    ) or issubclass(dtype, string)


# TODO: PY-632: unify these functions with others
def get_source_base_type(value):
    if inspect.isclass(value):
        value_type = value
        if not is_type_basic(value_type):
            warnings.warn('Setting schema with complex types is deprecated,'
                          ' use basic type instead', FutureWarning, stacklevel=2)
    else:
        warnings.warn('Setting schema with instance of the class is deprecated,'
                      ' use type instead', FutureWarning, stacklevel=2)
        value_type = type(value)
        # convert string to custom string if necessary
        if value_type is str and len(value) > string.DEFAULT_LENGTH:
            value_type = string[len(value)]

    if issubclass(value_type, bool):
        value_type = float

    if is_time_type(value_type):
        value_type = nsectime

    # check valid value type
    if get_base_type(value_type) not in [int, float, str, bool, decimal]:
        raise TypeError(f'Type "{repr(value_type)}" is not supported.')

    if not is_type_basic(value_type):
        raise TypeError(f"Type {repr(value_type)} can't be set in schema.")
    return value_type


def is_type_supported(dtype):
    return get_base_type(dtype) in [int, float, str, bool, decimal] or issubclass(dtype, (datetime, date))


def get_base_type(obj):
    if issubclass(obj, str):
        return str
    elif issubclass(obj, bool):
        return bool
    elif issubclass(obj, int):
        return int
    elif issubclass(obj, float):
        return float
    elif issubclass(obj, decimal):
        return decimal

    return type(None)


def get_object_type(obj):
    if isinstance(obj, (_nan, _inf)):
        return float
    if isinstance(obj, Type):
        return obj
    if hasattr(obj, 'dtype'):
        dtype = obj.dtype
        if isinstance(dtype, np.dtype):
            return dtype.type
        return dtype
    if is_time_type(obj):
        return nsectime
    # pylint: disable-next=unidiomatic-typecheck
    if type(obj) is int and obj > long.MAX:
        # by default we use python's int (onetick's long) for numbers
        # in case the number is too big, let's use onetick's ulong
        return ulong
    return type(obj)


def get_type_by_objects(objs):
    """
    Helper that calculates the widest type of the list passed objects.
    Used to determine type by returned values.
    """

    # collect types
    types = set()
    for v in objs:
        t = get_object_type(v)
        if issubclass(t, str):
            t = str
        types.add(t)

    # does not allow to mix string and numeric types
    dtype = None
    if str in types and (float in types or int in types or bool in types or nsectime in types or msectime in types):
        raise TypeError("It is not allowed to return values of string type and numeric type in one function.")

    # if there is only one value there, then
    # use it as is
    if len(types) == 1:
        dtype = next(iter(types))
        if dtype is bool:
            return dtype

    # process numeric types: the most generic is float
    if int in types:
        dtype = int
    if bool in types:
        dtype = float
    # None is equal to otp.nan
    if float in types or type(None) in types:
        dtype = float

    # process string types, taking into account OneTick long strings
    if str in types:
        max_len = string.DEFAULT_LENGTH
        for v in objs:
            t = get_object_type(v)
            if issubclass(t, string):
                if t.length is Ellipsis or max_len is Ellipsis:
                    max_len = Ellipsis
                else:
                    max_len = max(t.length, max_len)
            elif isinstance(v, str):
                max_len = max(len(v), max_len)

        if max_len == string.DEFAULT_LENGTH:
            dtype = str
        else:
            dtype = string[max_len]  # pylint: disable=E1136

    # process msectime and nsectime
    if dtype is float and (msectime in types or nsectime in types):
        raise TypeError("It is not allowed to return value of time type and float type in one function.")

    if msectime in types:
        dtype = msectime
    if nsectime in types:
        dtype = nsectime

    # we assume the None value has float default value, ie NaN
    # pylint: disable-next=unidiomatic-typecheck
    if type(None) is dtype:
        dtype = float

    return dtype


# ------------------- #
# extend datetime


class AbstractTime:
    def __init__(self):
        self.ts: pd.Timestamp

    @property
    def year(self):
        return self.ts.year

    @property
    def month(self):
        return self.ts.month

    @property
    def day(self):
        return self.ts.day

    def date(self):
        return _date(self.year, self.month, self.day)

    @property
    def start(self):
        return pd.Timestamp(self.year, self.month, self.day)

    @property
    def end(self):
        return pd.Timestamp(next_day(self.start))

    def strftime(self, fmt):
        return self.ts.strftime(fmt)

    @property
    def value(self):
        return self.ts.value

    def timestamp(self):
        return self.ts.timestamp()

    def __eq__(self, other):
        other = getattr(other, "ts", other)
        return self.ts == other

    def __hash__(self):
        return hash(self.ts)

    def __gt__(self, other):
        other = getattr(other, "ts", other)
        return self.ts > other

    def __ge__(self, other):
        other = getattr(other, "ts", other)
        return self.ts >= other

    def __lt__(self, other):
        other = getattr(other, "ts", other)
        return self.ts < other

    def __le__(self, other):
        other = getattr(other, "ts", other)
        return self.ts <= other


class datetime(AbstractTime):
    """
    Class :py:class:`otp.datetime <onetick.py.datetime>` is used for representing date with time in onetick-py.
    It can be used both when specifying start and end time for queries and
    in column operations with :py:class:`onetick.py.Source`.
    :ref:`Datetime offset objects <datetime_offsets>` (e.g. `otp.Nano`, `otp.Day`)
    can be added to or subtracted from `otp.datetime` object.

    Note
    ----
    Class :py:class:`otp.datetime <onetick.py.datetime>` share many methods
    that classes :pandas:`pandas.Timestamp` and :py:class:`datetime.datetime` have,
    but these objects are not fully interchangeable.
    Class :py:class:`otp.datetime <onetick.py.datetime>` should work in all onetick-py methods and classes,
    other classes should work too if documented,
    and may even work when not documented, but the users should not count on it.

    Parameters
    ----------
    first_arg: int, str, :py:class:`otp.datetime <onetick.py.datetime>`, :py:class:`otp.date <onetick.py.date>`\
                :pandas:`pandas.Timestamp`, :py:class:`datetime.datetime`
        If `month`, `day` and other parts of date are specified,
        first argument will be considered as year.
        Otherwise, first argument will be converted to :py:class:`otp.datetime <onetick.py.datetime>`.
    month: int
        Number between 1 and 12.
    day: int
        Number between 1 and 31.
    hour: int, default=0
        Number between 0 and 23.
    minute: int, default=0
        Number between 0 and 59.
    second: int, default=0
        Number between 0 and 59.
    microsecond: int, default=0
        Number between 0 and 999999.
    nanosecond: int, default=0
        Number between 0 and 999.
    tzinfo: :py:class:`datetime.tzinfo`
        Timezone object.
    tz: str
        Timezone name.

    Examples
    --------

    Initialization by :py:class:`datetime.datetime` class from standard library:

    >>> otp.datetime(datetime.datetime(2019, 1, 1, 1))
    2019-01-01 01:00:00

    Initialization by :pandas:`pandas.Timestamp` class:

    >>> otp.datetime(pd.Timestamp(2019, 1, 1, 1))
    2019-01-01 01:00:00

    Initialization by int timestamp:

    >>> otp.datetime(1234567890)
    1970-01-01 00:00:01.234567890

    Initialization by params with nanoseconds:

    >>> otp.datetime(2019, 1, 1, 1, 2, 3, 4, 5)
    2019-01-01 01:02:03.000004005

    Initialization by string:

    >>> otp.datetime('2019/01/01 1:02')
    2019-01-01 01:02:00

    `otp.dt` is the alias for `otp.datetime`:

    >>> otp.dt(2019, 1, 1)
    2019-01-01 00:00:00

    See also
    --------
    :ref:`Datetime offset objects <datetime_guide>`.
    """
    def __init__(
        self,
        first_arg,
        month=None,
        day=None,
        hour=None,
        minute=None,
        second=None,
        microsecond=None,
        nanosecond=None,
        *,
        tzinfo=None,
        tz=None,
    ):  # TODO: python 3.8 change first_arg to positional only arg
        tz, tzinfo = self._process_timezones_args(tz, tzinfo)

        if not any([month, day, hour, minute, second, microsecond, nanosecond]):
            result = self._create_from_one_arg(first_arg, tz, tzinfo)
        else:
            result = self._create_from_several_arg(first_arg, month, day, hour, minute, second, microsecond, nanosecond,
                                                   tzinfo)
        self.ts = result

    def _process_timezones_args(self, tz, tzinfo):
        if tz is not None:
            if tzinfo is None:
                # parameter tz in pandas.Timestamp is broken https://github.com/pandas-dev/pandas/issues/31929
                # it is fixed in pandas>=2.0.0, but we need to support older versions
                tzinfo = get_tzfile_by_name(tz)
                tz = None
            else:
                raise ValueError(
                    "tzinfo and tz params are mutually exclusive parameters, "
                    "they can't be specified both at the same time"
                )
        return tz, tzinfo

    def _create_from_several_arg(self, first_arg, month, day, hour, minute, second, microsecond, nanosecond, tzinfo):
        if nanosecond is not None and not (0 <= nanosecond <= 999):
            raise ValueError(
                "Nanosecond parameter should be between 0 and 999. "
                "Please use microsecond parameter or otp.Nano object."
            )
        if parse_version(pd.__version__) >= parse_version("2.0.0"):
            result = pd.Timestamp(
                first_arg, month, day, hour or 0, minute or 0, second or 0, microsecond or 0,
                nanosecond=nanosecond or 0,
            ).replace(tzinfo=tzinfo)
        else:
            result = pd.Timestamp(
                first_arg, month, day, hour or 0, minute or 0, second or 0, microsecond or 0, nanosecond or 0,
            ).replace(tzinfo=tzinfo)
        return result

    def _create_from_one_arg(self, first_arg, tz, tzinfo):
        arg_tz = getattr(first_arg, "tz", None)
        arg_tzinfo = getattr(first_arg, "tzinfo", None)
        if tz and arg_tz and arg_tz != tz or tzinfo and arg_tzinfo and arg_tzinfo != tzinfo:
            raise ValueError(
                "You've specified the timezone for the object, which already has it. "
                "It is recommended to swap the current timezone to desired by method of this object "
                "and then create otp.datetime object."
            )
        if isinstance(first_arg, (datetime, date)):
            first_arg = first_arg.ts
        kwargs = {}
        if tzinfo is not None:
            kwargs['tzinfo'] = tzinfo
        if tz is not None:
            kwargs['tz'] = tz
        result = pd.Timestamp(first_arg, **kwargs)
        return result

    @property
    def start(self):
        return super().start.replace(tzinfo=self.tzinfo)

    @property
    def end(self):
        return super().end.replace(tzinfo=self.tzinfo)

    def replace(self, **kwargs):
        """
        Replace parts of `otp.datetime` object.

        Parameters
        ----------
        year: int, optional
        month: int, optional
        day: int, optional
        hour: int, optional
        minute: int, optional
        second: int, optional
        microsecond: int, optional
        nanosecond: int, optional
        tzinfo: tz-convertible, optional

        Returns
        -------
        result: :py:class:`otp.datetime <onetick.py.datetime>`
            Timestamp with fields replaced.

        Examples
        --------
        >>> ts = otp.datetime(2022, 2, 24, 3, 15, 54, 999, 1)
        >>> ts
        2022-02-24 03:15:54.000999001
        >>> ts.replace(year=2000, month=2, day=2, hour=2, minute=2, second=2, microsecond=2, nanosecond=2)
        2000-02-02 02:02:02.000002002
        """
        return datetime(self.ts.replace(**kwargs))

    @property
    def tz(self):
        return self.ts.tz

    @property
    def tzinfo(self):
        return self.ts.tzinfo

    @property
    def hour(self):
        return self.ts.hour

    @property
    def minute(self):
        return self.ts.minute

    @property
    def second(self):
        return self.ts.second

    @property
    def microsecond(self):
        return self.ts.microsecond

    @property
    def nanosecond(self):
        return self.ts.nanosecond

    @staticmethod
    def now(tz=None):
        """
        Will return :py:class:`otp.datetime <onetick.py.datetime>` object
        with timestamp at the moment of calling this function.
        Not to be confused with function :func:`otp.now <onetick.py.now>` which can only add column
        with current timestamp to the :py:class:`otp.Source <onetick.py.Source>` when running the query.

        Parameters
        ----------
        tz : str or timezone object, default None
            Timezone to localize to.
        """
        return datetime(pd.Timestamp.now(tz))

    def __add__(self, other):
        """
        Add :ref:`datetime offset <datetime_offsets>` to otp.datetime.

        Parameters
        ----------
        other: :ref:`datetime offset <datetime_offsets>`, :py:class:`otp.datetime <onetick.py.datetime>`
            object to add

        Returns
        -------
        result: :py:class:`otp.datetime <onetick.py.datetime>`, :pandas:`pandas.Timedelta`
            return :py:class:`otp.datetime <onetick.py.datetime>`
            if otp.Nano or another :ref:`datetime offset <datetime_offsets>` object was passed as an argument,
            or :pandas:`pandas.Timedelta` object if :py:class:`otp.datetime <onetick.py.datetime>`
            was passed as an argument.

        Examples
        --------
        >>> otp.datetime(2022, 2, 24) + otp.Nano(1)
        2022-02-24 00:00:00.000000001
        """
        self._error_on_int_param(other, "+")
        return datetime(self.ts + other)

    def __sub__(self, other):
        """
        Subtract :ref:`datetime offset <datetime_offsets>` from otp.datetime.

        Parameters
        ----------
        other: :ref:`datetime offset <datetime_offsets>`, :py:class:`otp.datetime <onetick.py.datetime>`
            object to subtract

        Returns
        -------
        result: :py:class:`otp.datetime <onetick.py.datetime>`, :pandas:`pandas.Timedelta`
            return datetime if otp.Nano or another :ref:`datetime offset <datetime_offsets>`
            object was passed as an argument,
            or :pandas:`pandas.Timedelta` object if :py:class:`otp.datetime <onetick.py.datetime>`
            was passed as an argument.

        Examples
        --------
        >>> otp.datetime(2022, 2, 24) - otp.Nano(1)
        2022-02-23 23:59:59.999999999
        """
        self._error_on_int_param(other, "-")
        other = getattr(other, "ts", other)
        result = self.ts - other
        # do not convert to datetime in case timedelta is returned (arg is date)
        result = datetime(result) if isinstance(result, pd.Timestamp) else result
        return result

    def _error_on_int_param(self, other, op):
        if isinstance(other, int):
            raise TypeError(f"unsupported operand type(s) for {op}: 'otp.datetime' and 'int'")

    def __str__(self):
        return str(self.ts)

    def __repr__(self):
        return str(self.ts)

    def tz_localize(self, tz):
        """
        Localize tz-naive datetime object to a given timezone

        Parameters
        ----------
        tz: str or tzinfo
            timezone to localize datetime object into

        Returns
        -------
        result: :py:class:`otp.datetime <onetick.py.datetime>`
            localized datetime object

        Examples
        --------
        >>> d = otp.datetime(2021, 6, 3)
        >>> d.tz_localize("EST5EDT")
        2021-06-03 00:00:00-04:00
        """
        return datetime(self.ts.tz_localize(tz))

    def tz_convert(self, tz):
        """
        Convert tz-aware datetime object to another timezone

        Parameters
        ----------
        tz: str or tzinfo
            timezone to convert datetime object into

        Returns
        -------
        result: :py:class:`otp.datetime <onetick.py.datetime>`
            converted datetime object

        Examples
        --------
        >>> d = otp.datetime(2021, 6, 3, tz="EST5EDT")
        >>> d.tz_convert("Europe/Moscow")
        2021-06-03 07:00:00+03:00
        """
        return datetime(self.ts.tz_convert(tz))

    def to_operation(self, timezone=None):
        """
        Convert :py:class:`otp.datetime <onetick.py.datetime>` object to
        :py:class:`otp.Operation <onetick.py.Operation>`

        Parameters
        ----------
        timezone: Operation
            Can be used to specify timezone as an Operation.

        Examples
        --------
        >>> t = otp.Ticks(TZ=['EST5EDT', 'GMT'])
        >>> t['DT'] = otp.dt(2022, 1, 1).to_operation(timezone=t['TZ'])
        >>> otp.run(t, timezone='GMT')[['TZ', 'DT']]
                TZ                  DT
        0  EST5EDT 2022-01-01 05:00:00
        1      GMT 2022-01-01 00:00:00
        """
        return otp.Operation(op_str=otp.types.datetime2expr(self, timezone=timezone), dtype=otp.nsectime)


dt = datetime


class date(datetime):
    """
    Class ``date`` is used for representing date in onetick-py.
    It can be used both when specifying start and end time for queries and
    in column operations with :py:class:`onetick.py.Source`.

    Note
    ----
    Class :py:class:`otp.date <onetick.py.date>` share many methods
    that classes :pandas:`pandas.Timestamp` and :py:class:`datetime.date` have,
    but these objects are not fully interchangeable.
    Class :py:class:`otp.date <onetick.py.date>` should work in all onetick-py methods and classes,
    other classes should work too if documented,
    and may even work when not documented, but the users should not count on it.

    Parameters
    ----------
    first_arg: int, str, :py:class:`otp.datetime <onetick.py.datetime>`, :py:class:`otp.date <onetick.py.date>`\
                :pandas:`pandas.Timestamp`, :py:class:`datetime.datetime`, :py:class:`datetime.date`
        If `month` and `day` arguments are specified, first argument will be considered as year.
        Otherwise, first argument will be converted to otp.date.
    month: int
        Number between 1 and 12.
    day: int
        Number between 1 and 31.

    Examples
    --------
    :ref:`Datetime guide <datetime_guide>`.
    """

    def __init__(self, first_arg: Union[int, str, _date, _datetime, pd.Timestamp, AbstractTime],
                 month=None, day=None):
        if month is None and day is None:
            if isinstance(first_arg, AbstractTime):
                first_arg = first_arg.ts
            elif isinstance(first_arg, (int, str)):
                first_arg = pd.Timestamp(first_arg)
            if isinstance(first_arg, (_datetime, pd.Timestamp, datetime)):
                first_arg = first_arg.date()
            self.ts = pd.Timestamp(first_arg)  # remove hour, minutes and so on
        elif all((month, day)):
            self.ts = pd.Timestamp(first_arg, month, day)
        else:
            raise ValueError("Please specify three integers (year, month, day) "
                             "or object or create date from (string, int timestamp, "
                             "pandas.Timestamp, otp.datetime, otp.date, "
                             "datetime.datetime, datetime.date)")

    def __str__(self):
        return self.ts.strftime("%Y-%m-%d")

    def __repr__(self):
        return self.ts.strftime("%Y-%m-%d")

    def to_str(self, format="%Y%m%d"):
        """
        Convert date to string, by default it will be in YYYYMMDD format.

        Parameters
        ----------
        format: str
            strftime format of string to convert to.
        Returns
        -------
        result: str
        """
        return self.ts.strftime(format)


class _integer_meta(type):
    def __str__(cls):
        return getattr(cls, '_NAME', cls.__name__)

    @property
    def TYPE_SIZE(cls):
        return 8 * ctypes.sizeof(cls._CTYPE)

    @property
    def MIN(cls):
        if cls._UNSIGNED:
            return 0
        else:
            return -(2 ** (cls.TYPE_SIZE - 1))

    @property
    def MAX(cls):
        if cls._UNSIGNED:
            return (2 ** cls.TYPE_SIZE) - 1
        else:
            return (2 ** (cls.TYPE_SIZE - 1)) - 1


class _integer(int, metaclass=_integer_meta):
    def __new__(cls, value, *args, **kwargs):
        if not cls.MIN <= value <= cls.MAX:
            raise ValueError(f"{cls.__name__} values must be between {cls.MIN} and {cls.MAX}")
        return super().__new__(cls, value, *args, **kwargs)

    def __get_result(self, value):
        if isinstance(value, int):
            return self.__class__(self._CTYPE(value).value)
        return value

    def __add__(self, other):
        return self.__get_result(
            super().__add__(other)
        )

    def __radd__(self, other):
        return self.__get_result(
            super().__radd__(other)
        )

    def __sub__(self, other):
        return self.__get_result(
            super().__sub__(other)
        )

    def __rsub__(self, other):
        return self.__get_result(
            super().__rsub__(other)
        )

    def __mul__(self, other):
        return self.__get_result(
            super().__mul__(other)
        )

    def __rmul__(self, other):
        return self.__get_result(
            super().__rmul__(other)
        )

    def __truediv__(self, other):
        return self.__get_result(
            super().__truediv__(other)
        )

    def __rtruediv__(self, other):
        return self.__get_result(
            super().__rtruediv__(other)
        )

    def __str__(self):
        return super().__repr__()

    def __repr__(self):
        return f"{self.__class__.__name__}({self})"


class long(_integer):
    """
    OneTick data type representing signed long integer.

    The size of the type is not specified and may vary across different systems.
    Most commonly it's a 8-byte type with allowed values from -2**63 to 2**63 - 1.

    Note that the value is checked to be valid in constructor,
    but no overflow checking is done when arithmetic operations are performed.

    Examples
    --------
    >>> t = otp.Tick(A=otp.long(1))
    >>> t['B'] = otp.long(1) + 1
    >>> t.schema
    {'A': <class 'onetick.py.types.long'>, 'B': <class 'onetick.py.types.long'>}
    """
    _CTYPE = ctypes.c_long
    _UNSIGNED = False


class ulong(_integer):
    """
    OneTick data type representing unsigned long integer.

    The size of the type is not specified and may vary across different systems.
    Most commonly it's a 8-byte type with allowed values from 0 to 2**64 - 1.

    Note that the value is checked to be valid in constructor,
    but no overflow checking is done when arithmetic operations are performed.

    Examples
    --------
    >>> t = otp.Tick(A=otp.ulong(1))
    >>> t['B'] = otp.ulong(1) + 1
    >>> t.schema
    {'A': <class 'onetick.py.types.ulong'>, 'B': <class 'onetick.py.types.ulong'>}

    Note that arithmetic operations may result in overflow.
    Here we get 2**64 - 1 instead of -1.

    .. testcode::
       :skipif: is_windows()

       t = otp.Tick(A=otp.ulong(0) - 1)
       df = otp.run(t)
       print(df)

    .. testoutput::

               Time                     A
       0 2003-12-01  18446744073709551615
    """
    _CTYPE = ctypes.c_ulong
    _UNSIGNED = True


class _int(_integer):
    """
    OneTick data type representing signed integer.

    The size of the type is not specified and may vary across different systems.
    Most commonly it's a 4-byte type with allowed values from -2**31 to 2**31 - 1.

    Note that the value is checked to be valid in constructor,
    but no overflow checking is done when arithmetic operations are performed.

    Examples
    --------
    >>> t = otp.Tick(A=otp.int(1))
    >>> t['B'] = otp.int(1) + 1
    >>> t.schema
    {'A': <class 'onetick.py.types._int'>, 'B': <class 'onetick.py.types._int'>}
    """
    _CTYPE = ctypes.c_int
    _UNSIGNED = False
    _NAME = 'int'


class uint(_integer):
    """
    OneTick data type representing unsigned integer.

    The size of the type is not specified and may vary across different systems.
    Most commonly it's a 4-byte type with allowed values from 0 to 2**32 - 1.

    Note that the value is checked to be valid in constructor,
    but no overflow checking is done when arithmetic operations are performed.

    Examples
    --------
    >>> t = otp.Tick(A=otp.uint(1))
    >>> t['B'] = otp.uint(1) + 1
    >>> t.schema
    {'A': <class 'onetick.py.types.uint'>, 'B': <class 'onetick.py.types.uint'>}

    Note that arithmetic operations may result in overflow.
    Here we get 2**32 - 1 instead of -1.

    .. testcode::
       :skipif: is_windows()

       t = otp.Tick(A=otp.uint(0) - 1)
       df = otp.run(t)
       print(df)

    .. testoutput::

               Time           A
       0 2003-12-01  4294967295
    """
    _CTYPE = ctypes.c_uint
    _UNSIGNED = True


class byte(_integer):
    """
    OneTick data type representing byte integer.

    The size of the type is not specified and may vary across different systems.
    Most commonly it's a 1-byte type with allowed values from -128 to 127.

    Note that the value is checked to be valid in constructor,
    but no overflow checking is done when arithmetic operations are performed.

    Examples
    --------
    >>> t = otp.Tick(A=otp.byte(1))
    >>> t['B'] = otp.byte(1) + 1
    >>> t.schema
    {'A': <class 'onetick.py.types.byte'>, 'B': <class 'onetick.py.types.byte'>}

    Note that arithmetic operations may result in overflow.
    Here we get 127 instead of -129.

    >>> t = otp.Tick(A=otp.byte(-128) - 1)
    >>> otp.run(t)
            Time    A
    0 2003-12-01  127
    """
    _CTYPE = ctypes.c_byte
    _UNSIGNED = False


class short(_integer):
    """
    OneTick data type representing short integer.

    The size of the type is not specified and may vary across different systems.
    Most commonly it's a 2-byte type with allowed values from -32768 to 32767.

    Note that the value is checked to be valid in constructor,
    but no overflow checking is done when arithmetic operations are performed.

    Examples
    --------
    >>> t = otp.Tick(A=otp.short(1))
    >>> t['B'] = otp.short(1) + 1
    >>> t.schema
    {'A': <class 'onetick.py.types.short'>, 'B': <class 'onetick.py.types.short'>}

    Note that arithmetic operations may result in overflow.
    Here we get 32767 instead of -32769.

    >>> t = otp.Tick(A=otp.short(-32768) - 1)
    >>> otp.run(t)
            Time      A
    0 2003-12-01  32767
    """
    _CTYPE = ctypes.c_short
    _UNSIGNED = False


# ------------------- #
def type2str(t):
    if t is int:
        return "long"
    if t is str:
        return "string"
    if t is float:
        return "double"
    if t is None:
        return ''
    if t is decimal:
        return t._to_onetick_type_string()
    return str(t)


def str2type(type_name: str):
    """
    Converts OneTick type by its name into Python/OTP domain type.

    Parameters
    ----------
    type_name: str
        name of type from CSV or OneTick DB

    Returns
    -------
    class:
        Python/OTP type representing OneTick type
    """
    if type_name == "time32":
        return int
    if type_name == "long":
        # otp.long can be used too, but we use int for backward compatibility
        return int
    if type_name == "int":
        return _int
    if type_name == "byte":
        return byte
    if type_name == "short":
        return short
    if type_name == "uint":
        return uint
    if type_name == "ulong":
        return ulong
    elif type_name in ["double", "float"]:
        return float
    elif type_name == "decimal":
        return decimal
    elif type_name == "msectime":
        return msectime
    elif type_name == "nsectime":
        return nsectime
    elif type_name in ["string", "matrix", f"string[{string.DEFAULT_LENGTH}]"]:
        return str
    elif type_name == "varstring":
        return varstring
    elif type_name.find("string") != -1:
        length = int(type_name[type_name.find("[") + 1:type_name.find("]")])
        return string[length]
    return None


def type2np(t):
    if issubclass(t, str):
        if issubclass(t, otp.string) and isinstance(t.length, int):
            return '<U' + str(t.length)
        return '<U64'
    elif issubclass(t, bool):
        return 'boolean'
    elif issubclass(t, nsectime):
        return 'datetime64[ns]'
    elif issubclass(t, msectime):
        return 'datetime64[ms]'
    elif issubclass(t, datetime):
        return 'datetime64[ns]'
    elif issubclass(t, byte):
        if otq.webapi:
            return 'int8'
        return 'int32'
    elif issubclass(t, short):
        if otq.webapi:
            return 'int16'
        return 'int32'
    elif issubclass(t, uint):
        return 'uint32'
    elif issubclass(t, ulong):
        return 'uint64'
    elif issubclass(t, _int):
        return 'int32'
    elif issubclass(t, long):
        return 'int64'
    elif issubclass(t, int):
        return 'int64'
    elif issubclass(t, float):
        return 'float64'
    elif issubclass(t, decimal):
        return 'float64'
    else:
        return np.dtype(t)


def np2type(t):
    if not isinstance(t, (np.dtype, pd.api.extensions.ExtensionDtype)):
        raise ValueError(f'Unsupported value passed to `np2type`: `{t}` not numpy dtype')

    if t.name == 'int64':
        return int
    elif t.name == 'int8':
        return byte
    elif t.name == 'int16':
        return short
    elif t.name == 'int32':
        return _int
    elif t.name == 'uint64':
        return ulong
    elif t.name == 'uint32':
        return uint
    elif t.name == 'float64':
        return float
    elif t.name == 'boolean':
        return bool
    elif t.name.startswith('datetime64[ns'):
        return nsectime
    elif t.name.startswith('datetime64[us'):
        return nsectime
    elif t.name.startswith('datetime64[ms'):
        return msectime
    elif t.name == 'object':
        return str
    elif t.name == 'str':
        return str
    elif t.str.startswith('<U'):
        length = t.name[2:]
        if length:
            return string[int(length)]
        else:
            return str

    raise ValueError(f'Unknown numpy dtype passed to `np2type`: `{t}`')


# TODO: move this union of types to some common place
def datetime2expr(
    dt_obj: Union[_datetime, _date, pd.Timestamp, date, datetime],
    timezone: Optional[str] = None,
    timezone_naive: Optional[str] = None,
) -> str:
    """
    Convert python datetime values to OneTick string representation.
    If ``dt_obj`` is timezone-aware then timezone will be taken from ``dt_obj`` value.
    If ``dt_obj`` is timezone-naive then timezone specified with otp.config['tz'] or otp.run() will be used.

    Parameters
    ----------
    dt_obj
        date or datetime value
    timezone: str or Operation
        This timezone will be used unconditionally.
    timezone_naive: str or Operation
        This timezone will be used if ``dt_obj`` is timezone-naive.
    """
    dt_str = _format_datetime(dt_obj, '%Y-%m-%d %H:%M:%S.%f', add_nano_suffix=True)
    if timezone is None:
        timezone = get_timezone_from_datetime(dt_obj)
    if timezone is None:
        timezone = timezone_naive
    if not isinstance(timezone, otp.Operation):
        timezone = f'"{timezone}"' if timezone else '_TIMEZONE'
    return f'PARSE_NSECTIME("%Y-%m-%d %H:%M:%S.%J", "{dt_str}", {str(timezone)})'


def datetime2timeval(dt_obj: Union[_datetime, _date, pd.Timestamp, date, datetime], timezone: str = 'GMT'):
    """
    Get nanosecond-precision pyomd.timeval_t
    from different datetime types supported by onetick-py.

    If ``dt_obj`` is timezone-aware, then its timezone will be used.
    If ``dt_obj`` is timezone-naive , then it will be localized to ``timezone`` parameter (GMT by default).
    """
    dt_str = _format_datetime(dt_obj, '%Y-%m-%d %H:%M:%S.%f', add_nano_suffix=True)
    tz_str = get_timezone_from_datetime(dt_obj)
    if tz_str:
        # use timezone from the object
        return pyomd.TimeParser('%Y-%m-%d %H:%M:%S.%J', tz_str).parse_time(dt_str)
    else:
        # localize to the timezone specified in fucntion parameter
        return pyomd.TimeParser('%Y-%m-%d %H:%M:%S.%J', timezone).parse_time(dt_str)


def _format_datetime(dt_obj, fmt, add_nano_suffix=True):
    dt_str = dt_obj.strftime(fmt)
    if add_nano_suffix:
        if isinstance(dt_obj, (pd.Timestamp, datetime)):
            dt_str += f'{dt_obj.nanosecond:03}'[-3:]
        else:
            dt_str += '000'
    return dt_str


def value2str(v):
    """
    Converts a python value from the `v` parameter into OneTick format.
    """
    if issubclass(type(v), str):
        # there is no escape, so replacing double quotes with concatenation with it
        return '"' + str(v).replace('"', '''"+'"'+"''') + '"'

    if isinstance(v, decimal):
        return v._to_onetick_string()

    if isinstance(v, float) and not (isinstance(v, (_inf, _nan))):
        # PY-286: support science notation
        s = str(v)
        if "e" in s:
            return f'atof({value2str(s)})'
        if s == "nan":
            return str(nan)
        return s

    if is_time_type(v):
        return datetime2expr(v)

    if isinstance(v, nsectime):
        # we do not need the same for msectime because it works as is
        if int(v) == 0 or int(v) > 15e12:  # it is 2445/5/1
            return f'NSECTIME({v})'
        # This branch is for backward compatibility. Originally here was a bug that
        # allowed to pass only milliseconds as a value into the otp.nsectime constructor.
        # Obviously we expect there only nanoseconds, and the built-in NSECTIME works only
        # with nanoseconds.
        warnings.warn('It seems that you are using number of milliseconds as nanoseconds. ', stacklevel=2)

    return str(v)


# TODO: maybe can be removed, it is used only in tests now
def time2nsectime(time, timezone=None):
    """
    Converts complex time types to nsectime timestamp.

    Parameters
    ----------
    time: :py:class:`datetime.datetime`, :py:class:`datetime.date`,\
            :py:class:`otp.datetime <onetick.py.datetime>`, :py:class:`otp.date <onetick.py.date>`, pandas.Timestamp
        time to convert
    timezone:
        convert timezone before nsectime calculation

    Returns
    -------
    result: int
        number of nanoseconds since epoch
    """
    if isinstance(time, (_datetime, _date)):
        time = pd.Timestamp(time)
    elif isinstance(time, date):
        time = datetime(time)
    if timezone:
        if not has_timezone_parameter():  # accommodating legacy behavior prior to 20220327-3 weekly build
            time = time.replace(tzinfo=None)
        else:
            if time.tzinfo is None:
                time = time.tz_localize(timezone)
            else:
                # timezone conversion doesn't change the offset from epoch, only string representation,
                # so time.value (the number of nanoseconds) will stay the same
                # this line can be deleted
                time = time.tz_convert(timezone)
    return time.value


def is_time_type(time):
    """ Returns true if argument is subclass of any time type

    Checks if pass type is time type, currently checks for otp.date, otp.datetime,
    pd.Timestamp, datetime.date, datetime.datetime

    Parameters
    ----------
    time:
        object or type of the object

    Returns
    -------
    result: bool
        Return true if argument is time type

    Examples
    --------

    >>> is_time_type(datetime.datetime)  # OTdirective: skip-example: ;
    True
    >>> is_time_type(type(5))   # OTdirective: skip-example: ;
    False
    >>> is_time_type(datetime.datetime(2019, 1, 1))  # OTdirective: snippet-name: types.is time;
    True
    """
    time = time if inspect.isclass(time) else type(time)
    # do not check for datetime.datetime and pd.Timestamp, because they are in the same hierarchy
    # datetime.date -> datetime.datetime -> pd.Timestamp, where `->` means base class
    return issubclass(time, (_date, datetime, date))


def next_day(dt_obj: Union[_date, _datetime, date, datetime, pd.Timestamp]) -> _datetime:
    """
    Return the start of the next day of ``dt_obj`` as timezone-naive :py:class:`datetime.datetime`.

    Next day in this case means simply incrementing day/month/year number,
    not adding 24 hours (which may return the same date on DST days).
    """
    return _datetime(dt_obj.year, dt_obj.month, dt_obj.day) + _timedelta(days=1)


def default_by_type(dtype):
    """
    Get default value by OneTick type.

    Parameters
    ----------
    dtype:
        one of onetick-py base types

    Examples
    --------
    >>> otp.default_by_type(float)
    nan
    >>> otp.default_by_type(otp.decimal)
    decimal(0)
    >>> otp.default_by_type(int)
    0
    >>> otp.default_by_type(otp.ulong)
    ulong(0)
    >>> otp.default_by_type(otp.uint)
    uint(0)
    >>> otp.default_by_type(otp.short)
    short(0)
    >>> otp.default_by_type(otp.byte)
    byte(0)
    >>> otp.default_by_type(otp.nsectime)
    nsectime(0)
    >>> otp.default_by_type(otp.msectime)
    msectime(0)
    >>> otp.default_by_type(str)
    ''
    >>> otp.default_by_type(otp.string)
    string('')
    >>> otp.default_by_type(otp.string[123])
    string[123]('')
    >>> otp.default_by_type(otp.varstring)
    varstring('')
    """
    # TODO: think if we want to treat bool as basic onetick type
    if dtype is bool:
        return 0
    if not is_type_basic(dtype):
        raise TypeError(f"Can't get default value for type: {dtype}")
    if issubclass(dtype, int):
        return dtype(0)
    if dtype is otp.decimal:
        return otp.decimal(0)
    if issubclass(dtype, float):
        return nan
    if issubclass(dtype, str):
        return dtype('')
    if issubclass(dtype, nsectime) or issubclass(dtype, msectime):
        return dtype(0)
    raise TypeError(f"Can't get default value for type: {dtype}")


class timedelta(pd.Timedelta):
    """
    The object representing the delta between timestamps.

    Parameters
    ----------
    value: :py:class:`otp.timedelta <onetick.py.timedelta>`, :pandas:`pandas.Timedelta`,\
           :py:class:`datetime.timedelta`, str, or int
        Initialize this object from other types of objects.
    kwargs:
        Dictionary of offset names and their values.
        Available offset names:
        *weeks*, *days*, *hours*, *minutes*, *seconds*,
        *milliseconds*, *microseconds*, *nanoseconds*.

    Examples
    --------

    Create :py:class:`otp.timedelta <onetick.py.timedelta>` from key-value arguments:

    >>> otp.timedelta(weeks=1, days=1, hours=1, minutes=1, seconds=1, milliseconds=1, microseconds=1, nanoseconds=1)
    timedelta('8 days 01:01:01.001001001')

    Create :py:class:`otp.timedelta <onetick.py.timedelta>` from different types of objects:

    >>> otp.timedelta(datetime.timedelta(days=2, hours=3))
    timedelta('2 days 03:00:00')

    >>> otp.timedelta('20 days 13:02:01.999777666')
    timedelta('20 days 13:02:01.999777666')

    Adding :py:class:`otp.timedelta <onetick.py.timedelta>` object to :py:class:`otp.datetime <onetick.py.datetime>`:

    >>> otp.datetime(2022, 1, 1, 1, 2, 3) + otp.timedelta(days=1, hours=1, minutes=1, seconds=1)
    2022-01-02 02:03:04

    Adding :py:class:`otp.timedelta <onetick.py.timedelta>` object to :py:class:`otp.date <onetick.py.date>`:

    >>> otp.date(2022, 1, 1) + otp.timedelta(weeks=1, nanoseconds=1)
    2022-01-08 00:00:00.000000001
    """

    def __repr__(self):
        return super().__repr__().lower()

    def _get_offset(self):
        return self.value, 'nanosecond'



import os
from locator_parser.io import PrintWriter, FileReader
from locator_parser.actions import GetAll
from locator_parser.locator import Location, parse_locator
from locator_parser.common import apply_actions

from .temp import TmpFile
from .types import default


def get_dbs_locations_from_locator(locator):
    action = GetAll()
    action.add_where(Location)

    apply_actions(parse_locator, FileReader(locator), PrintWriter(), [action])

    return map(lambda x: x.location, action.result)


def tmp_locator(clean_up=default, empty=False):
    import onetick.py as otp

    STUBS = {'COMMON'}
    default_db = otp.config.get('default_db')
    if default_db:
        STUBS.add(default_db)

    data = []
    data.append('<VERSION_INFO VERSION="2"/>')
    data.append("<DATABASES>")
    data.append("")

    if not empty:
        for stub_name in STUBS:
            data.append(
                f'<db ID="{stub_name}" symbology="{otp.config.default_symbology}" time_series_is_composite="YES">'
            )
            data.append("<locations>")
            day_boundary_tz = otp.config.get('tz')
            day_boundary_tz = f'day_boundary_tz="{day_boundary_tz}"' if day_boundary_tz else ''
            data.append(
                '<location access_method="file" location="/opt/one_market_data/one_tick/examples/data/demo_level1" '
                'start_time="20001201000000" end_time="20301031050000" '
                f'{day_boundary_tz}/>'
            )
            data.append("</locations>")
            data.append('<feed type="heartbeat_generator">')
            data.append('<options format="native" />')
            data.append('</feed>')
            data.append("</db>")
            data.append("")
            data.append("")

    data.append("</DATABASES>")

    data.append("<TICK_SERVERS>")
    data.append("</TICK_SERVERS>")
    data.append("<CEP_TICK_SERVERS>")
    data.append("</CEP_TICK_SERVERS>")
    data.append("<INCLUDES>")
    data.append("</INCLUDES>")

    tmp_file = new_tmp_locator(clean_up)

    with open(tmp_file, "w") as fout:
        fout.write("\n".join(data))

    return tmp_file


def new_tmp_locator(clean_up):
    if os.getenv('OTP_WEBAPI_TEST_MODE'):
        tmp_file = TmpFile(name="locator.default", clean_up=clean_up, force=True)
    else:
        tmp_file = TmpFile(suffix=".locator", clean_up=clean_up)
    return tmp_file


def empty_locator(clean_up=default):
    data = []
    data.append('<VERSION_INFO VERSION="2"/>')
    data.append("<DATABASES>")
    data.append("</DATABASES>")
    data.append("<TICK_SERVERS>")
    data.append("</TICK_SERVERS>")
    data.append("<CEP_TICK_SERVERS>")
    data.append("</CEP_TICK_SERVERS>")
    data.append("<INCLUDES>")
    data.append("</INCLUDES>")

    tmp_file = new_tmp_locator(clean_up)
    with open(tmp_file, "w") as fout:
        fout.write("\n".join(data))
    return tmp_file



import os

from onetick.py.otq import otli
from onetick.py.db import DB
from onetick.py.session import Session


class TmpSession:
    def __init__(self):
        if Session._instance is None:
            if "ONE_TICK_CONFIG" in os.environ:
                # use external session
                otli.OneTickLib()
                self.session = None
                self.session_is_owned = False
            else:
                self.session_is_owned = True
                self.session = Session()

        else:
            self.session_is_owned = False
            self.session = Session._instance
            self.temp_dbs = []

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.session:  # exclude case when session is set externally, ie using ONE_TICK_CONFIG
            if self.session_is_owned:
                self.session.close()
            else:
                for db in self.temp_dbs:
                    self.session.locator.remove(db)

    def __db_is_registered(self, db):
        base_db = str(db).split("//")[0]    # noqa
        return (base_db in self.session.databases) or (base_db in self.session.locator.databases)

    def use(self, db):
        if self.session is None and isinstance(db, str) and "ONE_TICK_CONFIG" in os.environ:
            # we assume that database is already in the locator, that is
            # managed externally by user, and we don't need to add it
            # into the locator
            return

        if not isinstance(db, (str, DB)):
            raise TypeError(
                "Only a DB object or a string can be passed as db into get_"
                f"functions working with databases. Instead, {db.__class__} was provided."
            )

        if isinstance(db, str) and not self.__db_is_registered(db):
            if self.session_is_owned:
                raise TypeError(
                    "When there is no active session, db argument must be a DB object, not a database name."
                )
            else:
                raise TypeError(f"We can not find passed database {db} in the session")

        if not self.__db_is_registered(db):
            self.session.locator.add(db)
            if not self.session_is_owned:
                self.temp_dbs.append(db)



from onetick.py import types as ott
from onetick.py.core.column_operations.accessors._accessor import _Accessor
from onetick.py.core.column_operations.base import _Operation


class _FloatAccessor(_Accessor):
    """
    Accessor for float (double in Onetick terminology) functions

    >>> data = otp.Ticks(X=[1.1, 1.2])
    >>> data["Y"] = data["X"].float.<function_name>()   # doctest: +SKIP
    """

    def str(self, length=10, precision=6):
        """
        Converts float to str.

        Converts number to string with given ``length`` and ``precision``.
        The specified ``length`` should be greater than or equal
        to the part of the number before the decimal point plus the number's sign (if any).

        If ``length`` is specified as an int, the method will return strings with ``length`` characters,
        if ``length`` is specified as a column, the method will return string default (64 characters) length.

        Parameters
        ----------
        length: Operation or int
            Length of the string.
        precision: Operation or int
            Number of symbols after dot.

        Returns
        -------
        result: Operation
            String representation of float value.

        Examples
        --------

        >>> data = otp.Ticks(X=[1, 2.17, 10.31861, 3.141593, otp.nan, otp.inf, -otp.inf])
        >>> # OTdirective: snippet-name: float operations.to string.constant precision;
        >>> data["Y"] = data["X"].float.str(15, 3)
        >>> otp.run(data)
                             Time          X       Y
        0 2003-12-01 00:00:00.000   1.000000   1.000
        1 2003-12-01 00:00:00.001   2.170000   2.170
        2 2003-12-01 00:00:00.002  10.318610  10.319
        3 2003-12-01 00:00:00.003   3.141593   3.142
        4 2003-12-01 00:00:00.004        NaN     nan
        5 2003-12-01 00:00:00.005        inf     inf
        6 2003-12-01 00:00:00.006       -inf    -inf

        Parameters ``length`` and ``precision`` can also be taken from the columns:

        >>> data = otp.Ticks(X=[1, 2.17, 10.31841, 3.141593],
        ...                  LENGTH=[2, 3, 4, 5],
        ...                  PRECISION=[5, 5, 3, 3])
        >>> # OTdirective: snippet-name: float operations.to string.precision from fields;
        >>> data["Y"] = data["X"].float.str(data["LENGTH"], data["PRECISION"])
        >>> otp.run(data)
                             Time          X  LENGTH  PRECISION      Y
        0 2003-12-01 00:00:00.000   1.000000       2          5      1
        1 2003-12-01 00:00:00.001   2.170000       3          5    2.2
        2 2003-12-01 00:00:00.002  10.318410       4          3   10.3
        3 2003-12-01 00:00:00.003   3.141593       5          3  3.142
        """
        dtype = ott.string[length] if isinstance(length, int) else str

        def formatter(column, _length, _precision):
            column = ott.value2str(column)
            _length = ott.value2str(_length)
            _precision = ott.value2str(_precision)
            str_expr = f"str({column}, {_length}, {_precision})"
            # BDS-478: str() function raises exception for nan and inf values, so converting them manually
            str_expr = f'CASE({column}, NAN(), "nan", INFINITY(), "inf", -INFINITY(), "-inf", {str_expr})'
            return str_expr

        return _FloatAccessor.Formatter(
            op_params=[self._base_column, length, precision],
            dtype=dtype,
            formatter=formatter,
        )

    def cmp(self, other, eps):
        """
        Compare two double values between themselves according to ``eps`` relative difference.

        This function returns 0 if column = other, 1 if column > other, and -1 if column < other.
        Two numbers are considered to be equal if both of them are NaN or
        ``abs(column - other) / (abs(column) + abs(other)) < eps``.
        In other words, ``eps`` represents a relative difference (percentage) between the two numbers,
        not an absolute difference.

        Parameters
        ----------
        other: Operation or float
            column or value to compare with
        eps: Operation or float
            column or value with relative difference

        Returns
        -------
        result: Operation
            0 if column == other, 1 if column > other, and -1 if column < other.

        See Also
        --------
        eq

        Examples
        --------

        >>> data = otp.Ticks(X=[1, 2.17, 10.31841, 3.141593, 6],
        ...                  OTHER=[1.01, 2.1, 10.32841, 3.14, 5],
        ...                  EPS=[0, 1, 0.1, 0.001, 0.001])
        >>> # OTdirective: snippet-name: float operations.approximate comparison.lt|eq|gt;
        >>> data["Y"] = data["X"].float.cmp(data["OTHER"], data["EPS"])
        >>> otp.run(data)
                             Time          X     OTHER    EPS    Y
        0 2003-12-01 00:00:00.000   1.000000   1.01000  0.000 -1.0
        1 2003-12-01 00:00:00.001   2.170000   2.10000  1.000  0.0
        2 2003-12-01 00:00:00.002  10.318410  10.32841  0.100  0.0
        3 2003-12-01 00:00:00.003   3.141593   3.14000  0.001  0.0
        4 2003-12-01 00:00:00.004   6.000000   5.00000  0.001  1.0
        """
        def formatter(column, _other, _eps):
            column = ott.value2str(column)
            _other = ott.value2str(_other)
            _eps = ott.value2str(_eps)
            return f"double_compare({column}, {_other}, {_eps})"

        return _FloatAccessor.Formatter(
            op_params=[self._base_column, other, eps],
            dtype=float,
            formatter=formatter,
        )

    def eq(self, other, delta):
        """
        Compare two double values between themselves according to ``delta`` relative difference.
        Calculated as ``abs(column - other) <= delta``.

        Parameters
        ----------
        other: Operation, float
            column or value to compare with
        delta: Operation, float
            column or value with relative difference

        See Also
        --------
        cmp

        Returns
        -------
            Operation

        Examples
        --------

        >>> data = otp.Ticks(X=[1, 2.17, 10.31841, 3.141593, 6],
        ...                  OTHER=[1.01, 2.1, 10.32841, 3.14, 5],
        ...                  DELTA=[0, 1, 0.1, 0.01, 0.001])
        >>> # OTdirective: snippet-name: float operations.approximate comparison.eq;
        >>> data["Y"] = (1 + data["X"] - 1).float.eq(data["OTHER"], data["DELTA"])
        >>> otp.run(data)
                             Time          X     OTHER  DELTA    Y
        0 2003-12-01 00:00:00.000   1.000000   1.01000  0.000  0.0
        1 2003-12-01 00:00:00.001   2.170000   2.10000  1.000  1.0
        2 2003-12-01 00:00:00.002  10.318410  10.32841  0.100  1.0
        3 2003-12-01 00:00:00.003   3.141593   3.14000  0.010  1.0
        4 2003-12-01 00:00:00.004   6.000000   5.00000  0.001  0.0
        """
        def formatter(column, _other, _delta):
            column = ott.value2str(column)
            _other = ott.value2str(_other)
            _delta = ott.value2str(_delta)
            return f"abs({column} - {_other}) <= {_delta}"

        return _FloatAccessor.Formatter(
            op_params=[self._base_column, other, delta],
            dtype=bool,
            formatter=formatter,
        )



import io
import os
import subprocess
import warnings
import dataclasses
from pathlib import Path
from dataclasses import dataclass
from typing import Optional, Union, Type, Tuple

import pandas as pd
import onetick.py as otp
from onetick.py.otq import otq
from onetick.py.backports import cache
from onetick.py.core import query_inspector
from . import adaptive


if otp.__webapi__ or os.getenv("OTP_SKIP_OTQ_VALIDATION"):
    MEASURE_PERF = None
else:
    # see http://solutions.pages.soltest.onetick.com/iac/onetick-server/MeasurePerf.html
    MEASURE_PERF = Path(otp.__one_tick_bin_dir__) / 'measure_perf.exe'


@cache
def _get_allocation_lib() -> Optional[str]:
    suffix = '.dll' if os.name == 'nt' else '.so'
    allocation_lib = Path(otp.__one_tick_bin_dir__) / ('liballocation_interceptors' + suffix)
    if not allocation_lib.exists():
        warnings.warn(f"Can't find file {allocation_lib}, memory statistics will not be calculated")
        return None
    if os.name == 'nt':
        withdll_exe = Path(otp.__one_tick_bin_dir__) / 'withdll.exe'
        if not withdll_exe.exists():
            warnings.warn(f"Can't find file {withdll_exe}, memory statistics will not be calculated")
            return None
        return f'{withdll_exe} -d:{allocation_lib}'
    else:
        return f'LD_PRELOAD={allocation_lib}'


def _run_measure_perf(otq_file: str, summary_file: str, context: Optional[str] = None):
    if otp.__webapi__:
        raise RuntimeError("Can't use measure_perf.exe in WebAPI mode.")
    if MEASURE_PERF is not None and not MEASURE_PERF.exists():  # noqa
        raise RuntimeError(f"File {MEASURE_PERF} doesn't exist, can't execute it.")
    allocation_lib = _get_allocation_lib()
    cmd = allocation_lib or ''
    if cmd:
        cmd += ' '
    cmd += f'{MEASURE_PERF} -otq_file {otq_file} -summary_file {summary_file}'
    if context:
        cmd += f' -context {context}'
    try:
        subprocess.run(cmd, shell=True, check=True)
    except subprocess.CalledProcessError:
        if os.name == 'nt' and allocation_lib and Path(str(summary_file)).exists():
            # withdll.exe on Windows returns strange exit codes, but runs successfully anyway
            return
        raise


def measure_perf(src_or_otq: Union['otp.Source', str],
                 summary_file: Union[str, 'otp.utils.TmpFile', None] = None,
                 context: Union[str, Type[adaptive], None] = adaptive) -> Tuple[str, Union[str, 'otp.utils.TmpFile']]:
    """
    Run **measure_perf.exe** tool on some .otq file or :py:class:`onetick.py.Source`.
    Result is saved in file ``summary_file``.
    If it is not set, then temporary :py:class:`onetick.py.utils.temp.TmpFile` is generated and returned.

    Parameters
    ----------
    src_or_otq: :py:class:`~onetick.py.Source` or str
            :py:class:`~onetick.py.Source` object or path to already existing .otq file.
    summary_file: str
        path to the resulting summary file.
        By default some temporary file name will be used.
    context: str
        context that will be used to run the query.

    Returns
    -------
    Returns tuple with the path to the generated query and path to the summary file.

    Examples
    --------
    >>> t = otp.Tick(A=1)
    >>> otq_file, summary_file = otp.perf.measure_perf(t)
    >>> with open(summary_file) as f:  # doctest: +ELLIPSIS
    ...    print(f.read())
    Running result of ...
    ...
    index,EP_name,tag,...
    ...
    """
    if isinstance(src_or_otq, otp.Source):
        otq_file = src_or_otq.to_otq()
    else:
        otq_file = src_or_otq
    if not summary_file:
        summary_file = otp.utils.TmpFile()
    context_str = otp.config.context if context is adaptive else context
    _run_measure_perf(otq_file, str(summary_file), context_str)  # type: ignore
    return otq_file, summary_file


@dataclass
class SummaryEntry:
    def __getitem__(self, item):
        """
        Get value of the entry field by name.
        """
        return getattr(self, item)

    def __setitem__(self, item, value):
        """
        Set value of the entry field by name.
        """
        return setattr(self, item, value)

    def asdict(self) -> dict:
        """
        Return entry as a dictionary of field names and their values.
        """
        return dataclasses.asdict(self)

    def __iter__(self):
        """
        Iterator that returns tuples with name and value of each field.
        """
        for k, v in self.asdict().items():
            yield k, v

    @classmethod
    @cache
    def fields(cls):
        """
        Get list of entries field objects.
        """
        return dataclasses.fields(cls)

    @classmethod
    @cache
    def field_names(cls):
        """
        Get list of entries field names.
        """
        return [field.name for field in cls.fields()]


@dataclass
class DebugSummaryEntry:
    #: internal stack info number to identify debug information
    stack_info: Optional[str] = None
    #: python traceback string to identify location of the python code that created OneTick's EP
    traceback: Optional[str] = None


@dataclass
class _OrdinarySummaryEntry(SummaryEntry):
    #: Sequential number of the EP in the report
    index: int
    #: Name of the EP
    EP_name: str
    #: EP full tag (scope will be added to the tag if there is any)
    tag: int
    #: Time elapsed for EP execution with its child nodes in microseconds
    running_time_with_children: int
    #: Individual time elapsed for EP execution in microseconds
    running_time: int
    #: Number of ticks processed by the EP
    processed_tick_events: int
    #: Number of tick descriptors processed by the EP
    processed_schema_events: int
    processed_timer_events: int
    #: Maximal number of ticks accumulated by the EP during query execution
    #: This field is calculated only for aggregations (for example, EPs with a sliding window or GROUP_BY).
    #: For all other EPs, it has the value of 0.
    max_accumulated_ticks_count: int
    #: For continuous queries, each EP measures the latency in microseconds for all the ticks it has propagated.
    #: The latency of a tick is considered to be the difference between
    #: tick propagation host time and the timestamp of the tick.
    #: The maximum value of this latency (calculated by the EP)
    #: is reported by measure_perf.exe in the summary of that EP.
    #:
    #: The latency is calculated neither for aggregations with BUCKET_TIME=BUCKET_START
    #: (as ticks are propagated by overwritten timestamps that are equal to the bucket start) nor for their child EPs.
    #: For such cases, the following max_introduced_latency
    #: special values indicate the reason why the maximum introduced latency was not calculated:
    #:
    #: * -3 indicates that the EP is the culprit for latency calculation interruption
    #: * -2 indicates that the latency calculation for the EP is turned off because
    #:   its source EP's max_introduced_latency is -3
    #: * -1 indicates that the query is non-continuous
    max_introduced_latency: int
    #: There are EPs (like PRESORT, Aggregations, and others)
    #: that are allowed to propagate received ticks with some delay.
    #: This flag indicates if the EP introduces delay.
    ep_introduces_delay_flag: int
    #: The amount of memory allocated by EP and its child nodes.
    allocated_memory_with_children: int
    #: The amount of memory allocated by EP.
    allocated_memory: int
    #: The amount of memory unreleased by EP. The usual cause of non-zero unreleased memory is EP's cached data.
    unreleased_memory_with_children: int
    #: The amount of memory unreleased by EP and its child nodes.
    #: The usual cause of non-zero unreleased memory is EP's and its child nodes' cached data.
    unreleased_memory: int
    #: Peak memory utilization introduced by EP and its child nodes.
    peak_allocated_memory: int


@dataclass
class OrdinarySummaryEntry(DebugSummaryEntry, _OrdinarySummaryEntry):
    """
    Data class for each line of ordinary performance summary.
    """
    pass


@dataclass
class _PresortSummaryEntry(SummaryEntry):
    #: Sequential number of the branch in PRESORT EPs summary section
    index: int
    #: Source EP name of combined PRESORT EP source branch for which the summary was reported
    presort_source_ep_name: str
    #: Combined PRESORT EP name
    presort_sink_ep_name: str
    #: Source EP tag of combined PRESORT EP source branch for which the summary was reported
    presort_source_ep_tag: int
    #: Combined PRESORT EP tag
    presort_sink_ep_tag: int
    #: Maximum accumulated ticks count by PRESORT for the located branch.
    max_accumulated_ticks_count: int


@dataclass
class PresortSummaryEntry(DebugSummaryEntry, _PresortSummaryEntry):
    """
    Data class for each line of PRESORT performance summary.
    """
    pass


@dataclass
class _CEPSummaryEntry(SummaryEntry):
    #: Sequential number of the root EP in the root EPs summary section
    index: int
    #: Root EP name for which summary is provided
    sink_ep_name: str
    #: Root EP tag for which summary is provided
    sink_ep_tag: int
    #: Mean of the latencies of all ticks passed through the node
    latencies_mean: float
    #: Standard deviation of the latencies of all ticks passed through the node
    latencies_standard_deviation: float
    #: Average slope of the linear regression function found by least squares method calculated for all latencies
    #: of all ticks passed through the root node.
    #: As mentioned earlier, the regression function can be considered as a function describing some relationship
    #: between two variables: tick latency and tick arrival timestamp.
    latencies_average_slope: float
    #: This is the average variance of ticks latencies from the computed linear regression function.
    latencies_variance_from_regression_line: float


@dataclass
class CEPSummaryEntry(DebugSummaryEntry, _CEPSummaryEntry):
    """
    Data class for each line of CEP performance summary.
    """
    pass


class PerformanceSummary:
    _entry_cls: Optional[Type[SummaryEntry]] = None
    _entry_key: Optional[str] = None
    _ep_name_field: Optional[str] = None

    def __init__(self, text: Optional[str]):
        #: text of the summary (csv format)
        self.text = text
        #: pandas.DataFrame from the data of the summary
        self.dataframe = pd.read_csv(io.StringIO(self.text)) if self.text else pd.DataFrame()
        #: list of corresponding entries objects
        self.entries = self.dataframe.to_dict('records')
        #: mapping of EP tags to corresponding entry objects
        self.entries_dict = {}
        if self._entry_cls is not None:
            self.entries = [self._entry_cls(**e) for e in self.entries]
        if self._entry_key is not None:
            self.entries_dict = {e[self._entry_key]: e for e in self.entries}

    def __iter__(self):
        """
        Iterate over list of summary :attr:`entries`.
        """
        yield from self.entries


class OrdinarySummary(PerformanceSummary):
    """
    This is the first section in the summary file containing the largest portion of the summary for graph nodes.
    """

    _entry_cls = OrdinarySummaryEntry
    _entry_key = 'tag'
    _ep_name_field = 'EP_name'


class PresortSummary(PerformanceSummary):
    """
    In PRESORT EPs summary section **measure_perf.exe** provides per PRESORT source branch report
    containing max accumulated ticks count by PRESORT for each of these branches.
    Namely, it shows how many ticks were accumulated by PRESORT for each of these source branches.

    Please note that there are some PRESORT EP types, like SYNCHRONIZE_TIME EP,
    that do not support performance measurement, yet.

    Each line of this section contains six fields
    representing the location of the branch for which the report is printed
    and a field that contains the maximum number of ticks accumulated by PRESORT for this branch.

    The location of a branch is determined by the source and sink EP names and tags.
    """

    _entry_cls = PresortSummaryEntry
    _entry_key = 'presort_sink_ep_tag'
    _ep_name_field = 'presort_sink_ep_name'


class CEPSummary(PerformanceSummary):
    """
    The last summary type produced by **measure_perf.exe**
    is the latency summary for root EPs of the executed top-level query in CEP mode.

    Each root EP in CEP mode measures tick arrival latency before processing and propagating it to the sinks,
    down by the graph.

    Note that for non-CEP mode this summary is not printed at all.

    The summary provided in this section tries to shed some light
    and estimate the relationship between the following two variables:

    * dependent variable - tick latency
    * independent variable - tick arrival time into the root node.

    The summary printed in this section tries to describe this relationship using some statistical analysis metrics.

    Please note that these values are calculated across all ticks in all symbols processed by the query.

    Calculated stats for ROOT EPs are printed once the query is finished and there are no more ticks left to arrive.

    This summary contains the mean of latencies, standard deviation,
    average slope of linear regression function (calculated by the least squares method),
    and average variance from the regression function computed based on latency numbers of ticks
    that are passed through each root EP of a top-level query.

    For each root node, one line is printed with the fields containing values for each of the above-mentioned metrics.
    This summary should be enough to determine slow consumer queries and try to debug and optimize those.
    """

    _entry_cls = CEPSummaryEntry
    _entry_key = 'sink_ep_tag'
    _ep_name_field = 'sink_ep_name'


class PerformanceSummaryFile:
    def __init__(self, summary_file: Union[str, os.PathLike]):
        """
        Class to read and parse ``summary_file`` that was generated by OneTick's measure_perf.exe

        Parsed result is accessible via public properties of the class.

        Parameters
        ----------
        summary_file:
            path to the summary file.

        Examples
        --------
        >>> t = otp.Tick(A=1)
        >>> otq_file, summary_file = otp.perf.measure_perf(t)
        >>> result = otp.perf.PerformanceSummaryFile(summary_file)
        >>> print(result.ordinary_summary.dataframe)  # doctest: +ELLIPSIS
        index      EP_name  tag ...
            0  PASSTHROUGH    0 ...
        ...
        """
        #: path to the summary file
        self.summary_file = Path(summary_file)
        #: the text of the summary file
        self.summary_text = self.summary_file.read_text()
        ordinary_summary, presort_summary, cep_summary = self._parse()
        #: :class:`Ordinary summary <onetick.py.perf.OrdinarySummary>`
        self.ordinary_summary = ordinary_summary
        #: :class:`Presort summary <onetick.py.perf.PresortSummary>`
        self.presort_summary = presort_summary
        #: :class:`CEP summary <onetick.py.perf.CEPSummary>`
        self.cep_summary = cep_summary

    def _parse(self):
        summary_text_lines = self.summary_text.splitlines(keepends=True)

        summaries = {}
        for i, line in enumerate(summary_text_lines):
            if not line.startswith('index,'):
                continue
            header = set(line.strip().split(','))
            for summary_cls in (OrdinarySummary, PresortSummary, CEPSummary):
                if header.issubset(summary_cls._entry_cls.field_names()):
                    break
            else:
                raise ValueError("Can't parse performance summary")
            summaries[summary_cls] = i

        tables_header_line_indexes = list(summaries.values())
        for i, (summary_cls, start) in enumerate(summaries.items()):
            if i + 1 < len(tables_header_line_indexes):
                end = tables_header_line_indexes[i + 1]
            else:
                end = None
            summary_table_text = ''.join(summary_text_lines[start:end])
            summaries[summary_cls] = summary_cls(summary_table_text)

        return tuple(
            summaries.get(summary_cls) or summary_cls(None)
            for summary_cls in (OrdinarySummary, PresortSummary, CEPSummary)
        )


def _get_query_nodes(otq_file: str):
    """
    Get query nodes ids and stack info.
    """
    otq_file, _, query_name = otq_file.partition('::')
    info = query_inspector.get_query_info(otq_file, query_name)
    result = {}
    for node_tag, node in info.nodes.items():
        ep_name, stack_info = node._get_ep_name_and_stack_info()
        result[node_tag] = {'node_tag': node_tag, 'ep_name': ep_name, 'stack_info': stack_info}
    return result


class MeasurePerformance(PerformanceSummaryFile):
    def __init__(self, src_or_otq, summary_file=None, context=adaptive):
        """
        Class to run OneTick's measure_perf.exe on the specified query and parse the result.

        Additionally some debug information about the python location of event processor objects
        may be added to the result if
        :py:attr:`stack_info<onetick.py.configuration.Config.stack_info>`
        configuration parameter is set.

        Parsed result is accessible via public properties of the class.

        Parameters
        ----------
        src_or_otq: :py:class:`~onetick.py.Source` or str
            :py:class:`~onetick.py.Source` object or path to already existing .otq file.
        summary_file: str
            path to the resulting summary file.
            By default some temporary file name will be used.
        context: str
            context that will be used to run the query.

        Examples
        --------
        >>> t = otp.Tick(A=1)
        >>> result = otp.perf.MeasurePerformance(t)
        >>> print(result.ordinary_summary.dataframe)  # doctest: +ELLIPSIS
        index      EP_name  tag ...
            0  PASSTHROUGH    0 ...
        ...
        """

        self.otq_file, summary_file = measure_perf(src_or_otq, summary_file, context)
        super().__init__(summary_file)
        self._query_nodes = _get_query_nodes(self.otq_file)
        for summary in (self.ordinary_summary, self.presort_summary, self.cep_summary):
            self._add_debug_info_to_summary(summary)

    def _add_debug_info_to_summary(self, summary):
        """
        Modify entries in ``summary`` to include **stack_info** and **traceback** parameters.
        """
        from onetick.py._stack_info import _get_traceback_with_id
        for tag, entry in summary.entries_dict.items():
            if tag not in self._query_nodes:
                warnings.warn(f"Can't find node with tag {tag} in file {self.otq_file}")
                continue
            debug_info = self._query_nodes[tag]
            ep_name = entry[summary._ep_name_field]
            assert ep_name.split('/')[0] == debug_info['ep_name']
            stack_info_uuid = debug_info['stack_info']
            if stack_info_uuid:
                traceback = _get_traceback_with_id(stack_info_uuid)
                entry['stack_info'] = stack_info_uuid
                entry['traceback'] = traceback



import os
import re
import warnings
from typing import TYPE_CHECKING, Callable, Optional, Union

from onetick import py as otp
from onetick.py import types as ott
from onetick.py.core import db_constants
from onetick.py.core.lambda_object import _EmulateObject, apply_lambda, apply_script
from onetick.py.core.query_inspector import add_pins, get_query_info
from onetick.py.otq import otq

if TYPE_CHECKING:
    from onetick.py.core.source import Source


def apply_query(self: 'Source', query, in_pin="IN", out_pin="OUT", **params):
    """
    Apply data source to query

    .. deprecated:: 1.3.77

    See Also
    --------
    Source.apply
    """
    warnings.warn(
        'The "apply_query" method is deprecated. Please, use the .apply method instead or '
        "call a reference queries directly.",
        FutureWarning,
        stacklevel=2,
    )
    res = otp.apply_query(query, {in_pin: self}, [out_pin], **params)
    res.node().out_pin(out_pin)

    return res


def apply(self: 'Source', obj) -> Union['otp.Column', 'Source']:
    """
    Apply object to data source.

    Parameters
    ----------
    obj: onetick.py.query, Callable, type, onetick.query.GraphQuery

        - `onetick.py.query` allows to apply external nested query

        - python `Callable` allows to translate python code to similar OneTick's CASE expression.
            There are some limitations to which python operators can be used in this callable.
            See :ref:`Python callables parsing guide <python callable parser>` article for details.
            In :ref:`Remote OTP with Ray<ray-remote>` any `Callable` must be decorated with `@otp.remote` decorator,
            see :ref:`Ray usage examples<apply-remote-context>` for details.

        - `type` allows to apply default type conversion

        - `onetick.query.GraphQuery` allows to apply a build onetick.query.Graph

    Returns
    -------
    Column, Source

    Examples
    --------

    Apply external query to a tick flow. In this case it assumes that query has
    only one input and one output. Check the :class:`query` examples if you
    want to use a query with multiple inputs or outputs.

    >>> data = otp.Ticks(X=[1, 2, 3])
    >>> external_query = otp.query('update.otq')
    >>> data = data.apply(external_query)
    >>> otp.run(data)
                         Time  X
    0 2003-12-01 00:00:00.000  2
    1 2003-12-01 00:00:00.001  4
    2 2003-12-01 00:00:00.002  6

    Apply a predicate to a column / operation.
    In this case value passed to a predicate is column values.
    Result is a column.

    >>> data = otp.Ticks(X=[1, 2, 3])
    >>> data['Y'] = data['X'].apply(lambda x: x * 2)
    >>> otp.run(data)
                         Time  X  Y
    0 2003-12-01 00:00:00.000  1  2
    1 2003-12-01 00:00:00.001  2  4
    2 2003-12-01 00:00:00.002  3  6

    Another example of applying more sophisticated operation

    >>> data = otp.Ticks(X=[1, 2, 3])
    >>> data['Y'] = data['X'].apply(lambda x: 1 if x > 2 else 0)
    >>> otp.run(data)
                         Time  X  Y
    0 2003-12-01 00:00:00.000  1  0
    1 2003-12-01 00:00:00.001  2  0
    2 2003-12-01 00:00:00.002  3  1

    Example of applying a predicate to a Source. In this case value passed
    to a predicate is a whole tick. Result is a column.

    >>> data = otp.Ticks(X=[1, 2, 3], Y=[.5, -0.4, .2])
    >>> data['Z'] = data.apply(lambda tick: 1 if abs(tick['X'] * tick['Y']) > 0.5 else 0)
    >>> otp.run(data)
                         Time  X     Y  Z
    0 2003-12-01 00:00:00.000  1   0.5  0
    1 2003-12-01 00:00:00.001  2  -0.4  1
    2 2003-12-01 00:00:00.002  3   0.2  1

    See Also
    --------
    :py:class:`onetick.py.query`
    :py:meth:`onetick.py.Source.script`
    :ref:`Python callables parsing guide <python callable parser>`

    """
    if isinstance(obj, otp.query):
        graph = obj.graph_info
        if graph is None:
            return obj(IN=self)['OUT']

        if len(graph.nested_inputs) != 1:
            raise ValueError(
                f'It is expected the query "{obj.query_name}" to have one input, but it '
                f"has {len(graph.nested_inputs)}"
            )
        if len(graph.nested_outputs) > 1:
            raise ValueError(
                f'It is expected the query "{obj.query_name}" to have one or no output, but it '
                f"has {len(graph.nested_outputs)}"
            )

        in_pin = graph.nested_inputs[0].NESTED_INPUT

        if len(graph.nested_outputs) == 0:
            out_pin = None  # means no output
        else:
            out_pin = graph.nested_outputs[0].NESTED_OUTPUT

        return obj(**{in_pin: self})[out_pin]
    elif isinstance(obj, otq.GraphQuery):
        base_dir = None
        if os.getenv('OTP_WEBAPI_TEST_MODE'):
            from onetick.py.otq import _tmp_otq_path

            base_dir = _tmp_otq_path()
        tmp_file = otp.utils.TmpFile(suffix=".otq", base_dir=base_dir)

        obj.save_to_file(
            tmp_file.path,
            "graph_query_to_apply",
            # start and end times don't matter for this query, use some constants
            start=db_constants.DEFAULT_START_DATE,
            end=db_constants.DEFAULT_END_DATE,
        )

        query_info = get_query_info(tmp_file.path)

        if len(query_info.sources) != 1:
            raise ValueError(f"It is expected the query to have one input, but it has {len(query_info.sources)}")
        if len(query_info.sinks) != 1:
            raise ValueError(f"It is expected the query to have one output, but it has {len(query_info.sinks)}")

        add_pins(
            tmp_file.path,
            "graph_query_to_apply",
            [(query_info.sources[0], 1, "IN"), (query_info.sinks[0], 0, "OUT")],
        )

        return otp.query(tmp_file.path + "::graph_query_to_apply")(IN=self)["OUT"]

    return apply_lambda(obj, _EmulateObject(self))


def _extract_columns_from_script(script_text):
    result = {}
    supported_types = {
        'byte',
        'short',
        'uint',
        'long',
        'ulong',
        'int',
        'float',
        'double',
        'string',
        'time32',
        'nsectime',
        'msectime',
        'varstring',
        'decimal',
    }
    types = (
        r"(?P<type>varstring|byte|short|uint|int|ulong|long|"
        r"float|double|decimal|string|time32|msectime|nsectime|matrix)"
    )
    length = r"(\[(?P<length>\d+)\])?"
    name = r"\s+(?P<name>\w+)\s*=\s*"
    pattern = re.compile(types + length + name)
    for p in re.finditer(pattern, script_text):
        groupdict = p.groupdict()
        type_ = ott.str2type(groupdict["type"]) if groupdict["type"] in supported_types else None
        if type_:
            length = groupdict["length"]
            if length:
                length = int(length)
                if type_ is str and length != ott.string.DEFAULT_LENGTH:
                    type_ = ott.string[length]
            result[groupdict["name"]] = type_
        else:
            warnings.warn(
                f"{groupdict['type']} isn't supported for now, so field {groupdict['name']} won't "
                f"be added to schema."
            )
    return result


def script(self: 'Source', func: Union[Callable[['Source'], Optional[bool]], str], inplace=False) -> 'Source':
    # TODO: need to narrow Source object to get rid of undesired methods like aggregations
    """
    Implements a script for every tick.

    Allows to pass a ``func`` that will be applied per every tick.
    A ``func`` can be python callable in this case it will be translated to per tick script.
    In order to use it in Remote OTP with Ray, the function should be decorated with ``@otp.remote``,
    see :ref:`Ray usage examples<apply-remote-context>` for details.

    See :ref:`Per Tick Script Guide <python callable parser>` for more detailed description
    of python to OneTick code translation and per-tick script features.

    The script written in per tick script language can be passed itself as a string or path to a file with the code.
    onetick-py doesn't validate the script, but configure the schema accordingly.

    Parameters
    ----------
    func: callable, str or path
        - a callable that takes only one parameter - actual tick that behaves like a `Source` instance

        - or the script on per-tick script language

        - or a path to file with onetick script

    Returns
    -------
    :class:`Source`

    See also
    --------
    **PER_TICK_SCRIPT** OneTick event processor

    Examples
    --------
    >>> t = otp.Ticks({'X': [1, 2, 3], 'Y': [4, 5, 6]})
    >>> def fun(tick):
    ...     tick['Z'] = 0
    ...     if tick['X'] + tick['Y'] == 5:
    ...         tick['Z'] = 1
    ...     elif tick['X'] + tick['Y'] * 2 == 15:
    ...         tick['Z'] = 2
    >>> t = t.script(fun)
    >>> otp.run(t)
                         Time  X  Y  Z
    0 2003-12-01 00:00:00.000  1  4  1
    1 2003-12-01 00:00:00.001  2  5  0
    2 2003-12-01 00:00:00.002  3  6  2

    See also
    --------
    :py:meth:`onetick.py.Source.apply`
    :ref:`Per-Tick Script Guide <python callable parser>`
    """
    res = self if inplace else self.copy()
    changed_tick_lists = {}
    if callable(func):
        _new_columns, _script = apply_script(func, _EmulateObject(self))
        changed_tick_lists = _EmulateObject.get_changed_tick_lists()
    elif isinstance(func, str):
        if os.path.isfile(func):
            # path to the file with script
            with open(func) as file:
                _script = file.read()
        else:
            _script = func
        _new_columns = _extract_columns_from_script(_script)
    else:
        raise ValueError(
            "Wrong argument was specify, please use callable or string with either script on per tick "
            "language or path to it"
        )
    res.sink(otq.PerTickScript(script=_script))
    res.schema.update(**_new_columns)
    for tick_list_name, tick_list_schema in changed_tick_lists.items():
        res.state_vars[tick_list_name]._schema = tick_list_schema
    return res



from abc import ABCMeta, abstractmethod
from onetick.py.otq import otq
from onetick.py.backports import singledispatchmethod


class _BaseCutBuilder(metaclass=ABCMeta):
    """We need to build up query graph to compute bins based on max and min value.
    As we need to set proper output column name in per-tick script:
    1) instance of this class created on cut() and qcut() call,
    2) and then this instance is called during _Source.__setattr__() with output_column_name as argument.
    """

    def __init__(self, src, column, bins, **kwargs):
        self.src = src
        self.column = column
        self.bins = bins
        self.column_name = str(self.column)
        self.output_column_name = None
        self.labels = kwargs.get('labels')
        self.res = None
        self.bin_number = None
        if isinstance(self.bins, (int, float)) and self.labels and len(self.labels) != self.bins:
            raise ValueError('Number of labels is not equal to number of bins')

    def __call__(self, output_column_name):
        self.output_column_name = output_column_name
        self.build_graph()

    def build_graph(self):
        res = self.define_state_variables(self.bins)
        script = self.generate_script(res)
        self.src.script(script, inplace=True)

    @abstractmethod
    def compute_bin_variables(self):
        """Compute state variables based on min/max and number of bins/quantiles."""
        pass

    @singledispatchmethod
    def define_state_variables(self, bins):
        raise NotImplementedError(f'Method define_state_variables() is not implemented for parameter {type(bins)}')

    @define_state_variables.register
    def dsv_int(self, bins: int):
        res = self.compute_bin_variables()

        # merge src with bins res (to share state_vars)
        res.schema.set(**{})
        res, _ = res[(res['Time'] == 0)]
        self.src.sink(otq.Merge(identify_input_ts=False))
        self.src.source(res.node().copy_graph())
        self.src.node().add_rules(res.node().copy_rules())
        return res

    @define_state_variables.register
    def dsv_list(self, bins: list):
        return self.define_state_variables_by_list(bins)

    def define_state_variables_by_list(self, bins):
        raise NotImplementedError("define_state_variables_by_list() not implemented")

    def state_variable(self, inx):
        """State variable name used in bin's calculations"""
        return f'_TMP_{self.output_column_name}_{self.column_name}_{inx}'

    def generate_script(self, res):
        """
        Per-tick script generator.
        Every tick is compared state variable to find index of bin.
        Resulted bin (interval or label) is saved in output column.
        """
        bin_number = self.bin_number
        labels = self.labels
        column_name = self.column_name
        output_column_name = self.output_column_name

        s = ''
        for inx in range(1, bin_number):
            s += f'if ({column_name} <= STATE::{self.state_variable(inx)})' + '{\n'
            if labels is None:
                str_val = ''
                if inx > 1:
                    str_val = '"(" + ' + f'tostring(STATE::{self.state_variable(inx - 1)}, 10, 10) + '
                str_val += '", " + ' + f'tostring(STATE::{self.state_variable(inx)}, 10, 10)' + ' + "]"'
                s += f'{output_column_name} = ' + str_val + ';\n'
            else:
                s += f'{output_column_name} = "{labels[inx - 1]}";\n'

            s += '}'
            if inx < bin_number - 1:
                s += ' else '
            else:
                s += '\n'

        # default output column value is the last bin
        if labels is None:
            var_name = self.state_variable(bin_number - 1)
            self.src[output_column_name] = '(' + res.state_vars[var_name].apply(str) + ','
        else:
            self.src[output_column_name] = labels[-1]

        return s


# pylint: disable-next=abstract-method
class _CutBuilder(_BaseCutBuilder):

    def compute_bin_variables(self):
        """Compute state variables based on min/max and number of bins."""
        res = self.src.copy()
        column_name = self.column_name
        self.bin_number = self.bins

        res.sink(otq.Compute(
            compute=f"HIGH(INPUT_FIELD_NAME='{column_name}') HV,LOW(INPUT_FIELD_NAME='{column_name}') LV",
            append_output_field_name=False))
        res.schema.set(**{"HV": self.column.dtype, "LV": self.column.dtype})

        for inx in range(self.bins):
            res.state_vars[self.state_variable(inx)] = 0.
            res.state_vars[self.state_variable(inx)] = res['LV'] + (res['HV'] - res['LV']) / self.bins * inx
        return res

    def define_state_variables_by_list(self, bins: list):
        res = self.src
        self.bin_number = len(bins)
        for inx, value in enumerate(bins):
            res.state_vars[self.state_variable(inx)] = float(value)
        return res


class _QCutBuilder(_BaseCutBuilder):

    def compute_bin_variables(self):
        """Compute state variables based on otq.Percentile()."""
        res = self.src.copy()
        column_name = self.column_name
        self.bin_number = self.bins

        res.sink(otq.Percentile(number_of_quantiles=self.bins, input_field_names=column_name))
        res.sink(otq.Transpose(direction='ROWS_TO_COLUMNS', key_constraint_values=self.bins - 1))
        res.schema.set(**{column_name + f'_{inx}': self.column.dtype for inx in range(1, self.bins)})

        for inx in range(1, self.bins):
            res.state_vars[self.state_variable(inx)] = .0
            res.state_vars[self.state_variable(inx)] = res[f'{column_name}_{inx}']
        return res

    def define_state_variables_by_list(self, bins: list):
        raise NotImplementedError("qcut() with q as list of float is not implemented yet.")



import os

from .config import get_config_param


def abspath_to_query_by_otq_path(otq_path, query_path):
    """
    Function searches absolute path to a query based on the otq_path and
    short query path
    """
    query_path_parts = query_path.split("/")

    for base_path in otq_path.split(","):
        base_path = os.path.abspath(base_path)
        path = os.path.join(base_path, *query_path_parts)
        if os.path.exists(path):
            return path

    raise FileNotFoundError(f'Query "{query_path}" is not found')


def abspath_to_query_by_name(query_path):
    if os.path.isabs(query_path) and os.path.exists(query_path):
        return query_path

    if "ONE_TICK_CONFIG" not in os.environ:
        raise ValueError("ONE_TICK_CONFIG is not set!")

    return abspath_to_query_by_otq_path(
        get_config_param(os.environ["ONE_TICK_CONFIG"], "OTQ_FILE_PATH"), query_path
    )


def query_to_path_and_name(path):
    """
    Split passed OneTick like 'path' to a query ot the
    path and query name
    """
    query_path = None
    query_name = None

    pos1, pos2 = path.rfind(":"), path.find(":")
    if pos1 != pos2 and pos1 > 0 and pos2 > 0:
        _ = path.split(":")
        query_path, query_name = ":".join(_[:-2]), _[-1]
    else:
        query_path = path

    return query_path, query_name



from typing import TYPE_CHECKING, Optional

from onetick import py as otp
from onetick.py import types as ott
from onetick.py.otq import otq

from .misc import inplace_operation

if TYPE_CHECKING:
    from onetick.py.core.source import Source


@inplace_operation
def update_timestamp(
    self: 'Source',
    timestamp_field: str,
    timestamp_psec_field: Optional[str] = None,
    max_delay_of_original_timestamp=0,
    max_delay_of_new_timestamp=0,
    max_out_of_order_interval=0,
    max_delay_handling: str = 'complain',
    out_of_order_timestamp_handling: str = 'complain',
    zero_timestamp_handling: Optional[str] = None,
    log_sequence_violations: bool = False,
    inplace=False,
) -> Optional['Source']:
    """
    Assigns alternative timestamps to input ticks.

    In case resulting timestamps are out of order, time series is sorted in ascending order of those timestamps.

    Ticks with equal alternative timestamps are sorted in ascending order of the respective original timestamps,
    and equal ones among those are further sorted by ascending values of the **OMDSEQ** field,
    if such a field is present.

    Note
    ----

    In case parameters ``max_delay_of_original_timestamp`` or ``max_delay_of_new_timestamp`` are specified
    this method automatically uses :py:meth:`~onetick.py.Source.modify_query_times` method, affecting
    the start or end time of the query thus possibly changing some logic of the nodes placed higher in the graph.

    Also there are some limitations on using this method in the graph, e.g. this method can't be used in the
    "diamond" pattern and can't be used twice in the same graph.

    Parameters
    ----------
    timestamp_field: str
        Specifies the name of the input field, which is assumed to carry alternative timestamps.
    timestamp_psec_field: str
        Fractional (< 1 millisecond) parts of alternative timestamps will be taken from this field.
        They are assumed to be in picoseconds (0.001 of nanosecond).

        If this field is specified, then even if alternative timestamps have nanosecond granularity,
        only millisecond parts of them will be taken, nanosecond part will be rewritten.

        Useful in case ``timestamp_field`` has lower than millisecond granularity.

    max_delay_of_original_timestamp: int or :ref:`datetime offset<api/datetime/offsets/root:datetime offsets>`
        Changes the start time of the query to ``original_start_time - max_delay_of_original_timestamp``
        to make sure it processes all possible ticks that might have a new timestamp in the query time range.

        If integer value is specified, it is assumed to be milliseconds.

    max_delay_of_new_timestamp: int or :ref:`datetime offset<api/datetime/offsets/root:datetime offsets>`
        Changes the end time of the query to ``original_end_time + max_delay_of_new_timestamp``
        to make sure it processes all possible ticks that might have a new timestamp in the query time range.

        If integer value is specified, it is assumed to be milliseconds.

    max_out_of_order_interval: int or :ref:`datetime offset<api/datetime/offsets/root:datetime offsets>`
        Specifies the maximum out-of-order interval for alternative timestamps.
        Ticks with new timestamps out of order will be sorted in ascending order.

        This is the only parameter that leads to accumulation of ticks.

        If integer value is specified, it is assumed to be milliseconds.

    max_delay_handling: str ('complain' or 'discard' or 'use_original_timestamp' or 'use_new_timestamp')

        This parameter how to process ticks that are delayed even more than specified in
        ``max_delay_of_original_timestamp`` or ``max_delay_of_new_timestamp`` parameters.

        - **complain**: raise an exception

        - **discard**: do not add tick to the output time series

        - **use_original_timestamp**: assign original timestamp to the tick

        - **use_new_timestamp**: try to assign new timestamp to the tick. If the new timestamp of previous tick
            is greater than new timestamp of this tick then previous one is used.
            **NOTE: A previously propagated timestamp could be from a heartbeat.**
            When there is no previous propagated tick and the new timestamp falls behind the query start time,
            the latter is preferred, while query end time is preferred if the new timestamp exceeds it.

        This parameter is processed *before* any action from ``out_of_order_timestamp_handling`` parameter.

    out_of_order_timestamp_handling: str ('complain' or 'use_previous_value' or 'use_original_timestamp')

        This parameter how to process ticks that are out of order even more than specified in
        ``max_out_of_order_interval`` parameter.

        - **complain**: raise an exception

        - **use_previous_value**: assign new timestamp from a previous propagated tick

        - **use_original_timestamp**: try to use original timestamp. If maximum out-of-order interval is still
            exceeded then exception will be thrown.

        This parameter is processed *after* any action from ``max_delay_handling`` parameter.

    zero_timestamp_handling: str ('preserve_sequence'), optional

        This parameter specifies how to process ticks with zero alternative timestamps.

        If value is None, actions from ``max_delay_handling``
        and ``out_of_order_timestamp_handling`` are performed on it.

        If value is **preserve_sequence** then new timestamp will be set to the maximum between
        query start time and the new timestamp of the previous propagated tick.

    log_sequence_violations: bool
        If set to True, then warnings about actions from parameters ``out_of_order_timestamp_handling``,
        ``max_delay_handling`` and ``zero_timestamp_handling`` are logged into the log file.
    inplace: bool
        The flag controls whether operation should be applied inplace or not.
        If ``inplace=True``, then it returns nothing.
        Otherwise method returns a new modified object.

    Returns
    -------
    :class:`Source` or ``None``

    See also
    --------
    **UPDATE_TIMESTAMP** OneTick event processor

    Examples
    --------

    Data and timestamps from the database:

    >>> start = otp.dt(2022, 3, 2)
    >>> end = otp.dt(2022, 3, 3)
    >>> data = otp.DataSource('US_COMP', symbols='AAPL', tick_type='TRD')
    >>> otp.run(data, start=start, end=end)
                         Time  PRICE  SIZE
    0 2022-03-02 00:00:00.000    1.0   100
    1 2022-03-02 00:00:00.001    1.1   101
    2 2022-03-02 00:00:00.002    1.2   102

    Adding one hour to all ticks.
    Parameter ``max_delay_of_original_timestamp`` must be specified in this case:

    >>> data = otp.DataSource('US_COMP', symbols='AAPL', tick_type='TRD')
    >>> data['ORIG_TS'] = data['TIMESTAMP']
    >>> data['NEW_TS'] = data['TIMESTAMP'] + otp.Hour(1)
    >>> data = data.update_timestamp('NEW_TS', max_delay_of_original_timestamp=otp.Hour(1))
    >>> otp.run(data, start=start, end=end)[['Time', 'PRICE', 'SIZE', 'ORIG_TS']]
                         Time  PRICE  SIZE                  ORIG_TS
    0 2022-03-02 01:00:00.000    1.0   100  2022-03-02 00:00:00.000
    1 2022-03-02 01:00:00.001    1.1   101  2022-03-02 00:00:00.001
    2 2022-03-02 01:00:00.002    1.2   102  2022-03-02 00:00:00.002

    Subtracting one day from all ticks.
    Parameter ``max_delay_of_new_timestamp`` must be specified in this case.

    >>> data = otp.DataSource('US_COMP', symbols='AAPL', tick_type='TRD')
    >>> data['ORIG_TS'] = data['TIMESTAMP']
    >>> data['NEW_TS'] = data['TIMESTAMP'] - otp.Day(1)
    >>> data = data.update_timestamp('NEW_TS', max_delay_of_new_timestamp=otp.Day(1))
    >>> otp.run(data, start=start - otp.Day(1), end=end)[['Time', 'PRICE', 'SIZE', 'ORIG_TS']]
                         Time  PRICE  SIZE                 ORIG_TS
    0 2022-03-01 00:00:00.000    1.0   100 2022-03-02 00:00:00.000
    1 2022-03-01 00:00:00.001    1.1   101 2022-03-02 00:00:00.001
    2 2022-03-01 00:00:00.002    1.2   102 2022-03-02 00:00:00.002

    Parameter ``max_delay_handling`` can be used to specify how to handle ticks exceeding the maximum:

    >>> data = otp.DataSource('US_COMP', symbols='AAPL', tick_type='TRD')
    >>> data['ORIG_TS'] = data['TIMESTAMP']
    >>> data['NEW_TS'] = data.apply(
    ...     lambda row: row['TIMESTAMP'] + otp.Hour(24)
    ...     if row['PRICE'] == 1.1
    ...     else row['TIMESTAMP'] + otp.Hour(1)
    ... )
    >>> data = data.update_timestamp('NEW_TS',
    ...                              max_delay_of_original_timestamp=otp.Hour(1),
    ...                              max_delay_handling='discard')
    >>> otp.run(data, start=start, end=end)[['Time', 'PRICE', 'SIZE', 'ORIG_TS']]
                         Time  PRICE  SIZE                  ORIG_TS
    0 2022-03-02 01:00:00.000    1.0   100  2022-03-02 00:00:00.000
    1 2022-03-02 01:00:00.002    1.2   102  2022-03-02 00:00:00.002

    Parameter ``max_out_of_order_interval`` can be used in case new timestamp are out of order:

    >>> data = otp.DataSource('US_COMP', symbols='AAPL', tick_type='TRD')
    >>> data['ORIG_TS'] = data['TIMESTAMP']
    >>> data = data.agg({'COUNT': otp.agg.count()}, running=True, all_fields=True)
    >>> data['NEW_TS'] = data['TIMESTAMP'] - otp.Minute(data['COUNT'])
    >>> data = data.update_timestamp('NEW_TS',
    ...                              max_delay_of_new_timestamp=otp.Hour(10),
    ...                              max_out_of_order_interval=otp.Minute(100))
    >>> otp.run(data, start=start - otp.Hour(2), end=end)[['Time', 'PRICE', 'SIZE', 'ORIG_TS', 'COUNT']]
                         Time  PRICE  SIZE                 ORIG_TS  COUNT
    0 2022-03-01 23:57:00.002    1.2   102 2022-03-02 00:00:00.002      3
    1 2022-03-01 23:58:00.001    1.1   101 2022-03-02 00:00:00.001      2
    2 2022-03-01 23:59:00.000    1.0   100 2022-03-02 00:00:00.000      1
    """
    if timestamp_field not in self.schema:
        raise ValueError(f"Field '{timestamp_field}' is not in schema")
    if self.schema[timestamp_field] not in {otp.msectime, otp.nsectime}:
        raise ValueError(f"Unsupported type for 'timestamp_field': {self.schema[timestamp_field]}")

    if timestamp_psec_field is not None:
        if timestamp_psec_field not in self.schema:
            raise ValueError(f"Field '{timestamp_psec_field}' is not in schema")
        if self.schema[timestamp_psec_field] is not int:
            raise ValueError(f"Unsupported type for 'timestamp_psec_field': {self.schema[timestamp_psec_field]}")

    if max_delay_handling not in {'complain', 'discard', 'use_original_timestamp', 'use_new_timestamp'}:
        raise ValueError(f"Unsupported value for parameter 'max_delay_handling': {max_delay_handling}")

    if out_of_order_timestamp_handling not in {'complain', 'use_previous_value', 'use_original_timestamp'}:
        raise ValueError(
            f"Unsupported value for parameter 'out_of_order_timestamp_handling': {out_of_order_timestamp_handling}"
        )

    if zero_timestamp_handling not in {None, 'preserve_sequence'}:
        raise ValueError(f"Unsupported value for parameter 'zero_timestamp_handling': {zero_timestamp_handling}")

    def get_ms(value):
        if isinstance(value, int):
            return value
        try:
            nanos = value.nanos
        except Exception as e:
            raise ValueError(f'Unsupported parameter value: {type(value)}') from e
        if nanos < 1_000_000:
            raise ValueError('Values less than one millisecond are not supported in this parameter')
        return nanos // 1_000_000

    self.sink(
        otq.UpdateTimestamp(
            timestamp_field_name=timestamp_field,
            timestamp_psec_field_name=timestamp_psec_field or '',
            max_delay_of_original_timestamp=get_ms(max_delay_of_original_timestamp),
            max_delay_of_new_timestamp=get_ms(max_delay_of_new_timestamp),
            max_out_of_order_interval_msec=get_ms(max_out_of_order_interval),
            max_delay_handling=max_delay_handling,
            out_of_order_timestamp_handling=out_of_order_timestamp_handling,
            zero_timestamp_handling=zero_timestamp_handling,
            log_sequence_violations=log_sequence_violations,
        )
    )
    return self


@inplace_operation
def modify_query_times(
    self: 'Source', start=None, end=None, output_timestamp=None, propagate_heartbeats=True, inplace=False
) -> Optional['Source']:
    """
    Modify ``start`` and ``end`` time of the query.

    * query times are changed for all operations
        only **before** this method up to the source of the graph.
    * all ticks' timestamps produced by this method
        **must** fall into original time range of the query.

    It is possible to change ticks' timestamps with parameter ``output_timestamp``,
    so they will stay inside the original time range.

    Note
    ----
    Due to how OneTick works internally, tick generators
    :py:class:`otp.Tick <onetick.py.Tick>` and :py:func:`otp.Ticks <onetick.py.Ticks>`
    are not affected by this method.

    Parameters
    ----------
    start: :py:class:`otp.datetime <onetick.py.datetime>` or \
            :py:class:`~onetick.py.core.source.MetaFields` or :py:class:`~onetick.py.Operation`
        Expression to replace query start time.
        By default, start time is not changed.
        Note that expression in this parameter can't depend on ticks, thus only
        :py:class:`~onetick.py.core.source.MetaFields` and constants can be used.
    end: :py:class:`otp.datetime <onetick.py.datetime>` or \
            :py:class:`~onetick.py.core.source.MetaFields` or :py:class:`~onetick.py.Operation`
        Expression to replace query end time.
        By default, end time is not changed.
        Note that expression in this parameter can't depend on ticks, thus only
        :py:class:`~onetick.py.core.source.MetaFields` and constants can be used.
    output_timestamp: :py:class:`onetick.py.Operation`
        Expression that produces timestamp for each tick.
        By default, the following expression is used: ``orig_start + orig_timestamp - start``
        This expression covers cases when start time of the query is changed and keeps
        timestamp inside original time range.
        Note that it doesn't cover cases, for example, if end time was increased,
        you have to handle such cases yourself.
    propagate_heartbeats: bool
        Controls heartbeat propagation.
    inplace: bool
        The flag controls whether operation should be applied inplace or not.
        If ``inplace=True``, then it returns nothing.
        Otherwise method returns a new modified object.

    See also
    --------
    | **MODIFY_QUERY_TIMES** OneTick event processor
    | :py:meth:`onetick.py.Source.time_interval_shift`

    Returns
    -------
    :class:`Source` or ``None``

    Examples
    --------

    >>> start = otp.dt(2022, 3, 2)
    >>> end = otp.dt(2022, 3, 2) + otp.Milli(3)
    >>> data = otp.DataSource('US_COMP', symbols='AAPL', tick_type='TRD')

    By default, method does nothing:

    >>> t = data.modify_query_times()
    >>> otp.run(t, start=start, end=end)
                         Time  PRICE  SIZE
    0 2022-03-02 00:00:00.000    1.0   100
    1 2022-03-02 00:00:00.001    1.1   101
    2 2022-03-02 00:00:00.002    1.2   102

    See how ``_START_TIME`` and ``_END_TIME`` meta fields are changed.
    They are changed *before* ``modify_query_times``:

    >>> t = data.copy()
    >>> t['S_BEFORE'] = t['_START_TIME']
    >>> t['E_BEFORE'] = t['_END_TIME']
    >>> t = t.modify_query_times(start=t['_START_TIME'] + otp.Milli(1),
    ...                          end=t['_END_TIME'] - otp.Milli(1))
    >>> t['S_AFTER'] = t['_START_TIME']
    >>> t['E_AFTER'] = t['_END_TIME']
    >>> otp.run(t, start=start, end=end)
            Time  PRICE  SIZE                S_BEFORE                E_BEFORE    S_AFTER                 E_AFTER
    0 2022-03-02    1.1   101 2022-03-02 00:00:00.001 2022-03-02 00:00:00.002 2022-03-02 2022-03-02 00:00:00.003

    You can decrease time interval without problems:

    >>> t = data.modify_query_times(start=data['_START_TIME'] + otp.Milli(1),
    ...                             end=data['_END_TIME'] - otp.Milli(1))
    >>> otp.run(t, start=start, end=end)
            Time  PRICE  SIZE
    0 2022-03-02    1.1   101

    Note that the timestamp of the tick was changed with default expression.
    In this case we can output original timestamps,
    because they fall into original time range:

    >>> t = data.modify_query_times(start=data['_START_TIME'] + otp.Milli(1),
    ...                             end=data['_END_TIME'] - otp.Milli(1),
    ...                             output_timestamp=data['TIMESTAMP'])
    >>> otp.run(t, start=start, end=end)
                         Time  PRICE  SIZE
    0 2022-03-02 00:00:00.001    1.1   101

    But it will not work if new time range is wider than original:

    >>> t = data.modify_query_times(start=data['_START_TIME'] - otp.Milli(1),
    ...                             output_timestamp=data['TIMESTAMP'])
    >>> otp.run(t, start=start + otp.Milli(1), end=end + otp.Milli(1)) # doctest: +ELLIPSIS
    Traceback (most recent call last):
    Exception...timestamp is falling out of initial start/end time range...

    In this case default ``output_timestamp`` expression would work just fine:

    >>> t = data.modify_query_times(start=data['_START_TIME'] - otp.Milli(1))
    >>> otp.run(t, start=start + otp.Milli(1), end=end + otp.Milli(1))
                         Time  PRICE  SIZE
    0 2022-03-02 00:00:00.001    1.0   100
    1 2022-03-02 00:00:00.002    1.1   101
    2 2022-03-02 00:00:00.003    1.2   102

    But it doesn't work, for example, if end time has crossed the borders of original time range.
    In this case other ``output_timestamp`` expression must be specified:

    >>> t = data.modify_query_times(
    ...     start=data['_START_TIME'] - otp.Milli(2),
    ...     output_timestamp=otp.math.min(data['TIMESTAMP'] + otp.Milli(2), data['_END_TIME'])
    ... )
    >>> otp.run(t, start=start + otp.Milli(2), end=end)
                         Time  PRICE  SIZE
    0 2022-03-02 00:00:00.002    1.0   100
    1 2022-03-02 00:00:00.003    1.1   101
    2 2022-03-02 00:00:00.003    1.2   102

    Remember that ``start`` and ``end`` parameters can't depend on ticks:

    >>> t = data.copy()
    >>> t['X'] = 12345
    >>> t = t.modify_query_times(start=t['_START_TIME'] + t['X'] - t['X'],
    ...                          end=t['_END_TIME'] - otp.Milli(1))
    >>> otp.run(t, start=start, end=end) # doctest: +ELLIPSIS
    Traceback (most recent call last):
    Exception...parameter must not depend on ticks...

    Constant datetime values can be used as parameters too:

    >>> t = data.modify_query_times(start=start + otp.Milli(1),
    ...                             end=end - otp.Milli(1))
    >>> otp.run(t, start=start, end=end)
            Time  PRICE  SIZE
    0 2022-03-02    1.1   101

    Note that some graph patterns are not allowed when using this method.
    For example, modifying query times for a branch that will be merged later:

    >>> t1, t2 = data[data['PRICE'] > 1.3]
    >>> t2 = t2.modify_query_times(start=start + otp.Milli(1))
    >>> t = otp.merge([t1, t2])
    >>> otp.run(t, start=start, end=end) # doctest: +ELLIPSIS
    Traceback (most recent call last):
    Exception...Invalid graph...time bound to a node...an intermediate node in one of the cycles in graph...
    """
    start = ott.value2str(start) if start is not None else ''
    end = ott.value2str(end) if end is not None else ''
    output_timestamp = ott.value2str(output_timestamp) if output_timestamp is not None else ''
    self.sink(
        otq.ModifyQueryTimes(
            start_time=start,
            end_time=end,
            output_timestamp=output_timestamp,
            propagate_heartbeats=propagate_heartbeats,
        )
    )
    return self


def time_interval_shift(self: 'Source', shift, inplace=False) -> Optional['Source']:
    """
    Shifting time interval for a source.

    The whole data flow is shifted all the way up to the source of the graph.

    The start and end times of the query will be changed for all operations before this method,
    and will stay the same after this method.

    WARNING: The ticks' timestamps *are changed* automatically so they fit into original time range.

    You will get different set of ticks from the database, but the timestamps of the ticks
    from that database will not be the same as in the database.

    They need to be changed so they fit into the original query time range.
    See details in :py:meth:`onetick.py.Source.modify_query_times`.

    Parameters
    ----------
    shift: int or :ref:`datetime offset<api/datetime/offsets/root:datetime offsets>`
        Offset to shift the whole time interval.
        Can be positive or negative.
        Positive value moves time interval into the future, negative -- to the past.
        int values are interpreted as milliseconds.

        Timestamps of the ticks will be changed so they fit into the original query time range
        by subtracting ``shift`` from each timestamp.
    inplace: bool
        The flag controls whether operation should be applied inplace or not.
        If ``inplace=True``, then it returns nothing.
        Otherwise method returns a new modified object.

    Returns
    -------
    :class:`Source` or ``None``

    See also
    --------
    | :py:meth:`onetick.py.Source.modify_query_times`
    | :py:meth:`onetick.py.Source.time_interval_change`

    Examples
    --------

    --> Also see use-case using :py:meth:`time_interval_shift` for calculating
    `Markouts <../../static/getting_started/use_cases.html#point-in-time-benchmarks-bbo-at-different-markouts>`_


    >>> start = otp.dt(2022, 3, 2)
    >>> end = otp.dt(2022, 3, 2) + otp.Milli(3)
    >>> data = otp.DataSource('US_COMP', symbols='AAPL', tick_type='TRD')

    Default data:

    >>> otp.run(data, start=start, end=end)
                         Time  PRICE  SIZE
    0 2022-03-02 00:00:00.000    1.0   100
    1 2022-03-02 00:00:00.001    1.1   101
    2 2022-03-02 00:00:00.002    1.2   102

    Get window for a third tick:

    >>> otp.run(data, start=start + otp.Milli(2), end=start + otp.Milli(3))
                         Time  PRICE  SIZE
    0 2022-03-02 00:00:00.002    1.2   102

    Shifting time window will result in different set of ticks,
    but the ticks will have their timestamps changed to fit into original time range.
    Let's shift time 2 milliseconds back and thus get the first tick:

    >>> t = data.time_interval_shift(shift=-otp.Milli(2))
    >>> otp.run(t, start=start + otp.Milli(2), end=start + otp.Milli(3))
                         Time  PRICE  SIZE
    0 2022-03-02 00:00:00.002    1.0   100

    Here we are querying empty time interval, but shifting one second back to get ticks.

    >>> t = data.time_interval_shift(shift=-otp.Second(1))
    >>> otp.run(t, start=start + otp.Second(1), end=end + otp.Second(1))
                         Time  PRICE  SIZE
    0 2022-03-02 00:00:01.000    1.0   100
    1 2022-03-02 00:00:01.001    1.1   101
    2 2022-03-02 00:00:01.002    1.2   102

    Note that tick generators
    :py:class:`otp.Tick <onetick.py.Tick>` and :py:func:`otp.Ticks <onetick.py.Ticks>`
    are not really affected by this method, they will have the same timestamps:

    >>> t = otp.Tick(A=1)
    >>> otp.run(t)
            Time  A
    0 2003-12-01  1

    >>> t = t.time_interval_shift(shift=otp.Second(1))
    >>> otp.run(t)
            Time  A
    0 2003-12-01  1
    """
    start = self['_START_TIME'] + shift
    end = self['_END_TIME'] + shift
    # change timestamps so they fit into original time range
    output_timestamp = self['TIMESTAMP'] - shift
    return self.modify_query_times(start=start, end=end, output_timestamp=output_timestamp, inplace=inplace)


def time_interval_change(self: 'Source', start_change=0, end_change=0, inplace=False) -> Optional['Source']:
    """
    Changing time interval by making it bigger or smaller.

    All timestamps of ticks that are crossing the border of original time range
    will be set to original start time or end time depending on their original time.

    Parameters
    ----------
    start_change: int or :ref:`datetime offset <api/datetime/offsets/root:datetime offsets>`
        Offset to shift start time.
        Can be positive or negative.
        Positive value moves start time into the future, negative -- to the past.
        int values are interpreted as milliseconds.
    end_change: int or :ref:`datetime offset <api/datetime/offsets/root:datetime offsets>`
        Offset to shift end time.
        Can be positive or negative.
        Positive value moves end time into the future, negative -- to the past.
        int values are interpreted as milliseconds.
    inplace: bool
        The flag controls whether operation should be applied inplace or not.
        If ``inplace=True``, then it returns nothing.
        Otherwise method returns a new modified object.

    Returns
    -------
    :class:`Source` or ``None``

    See also
    --------
    | :py:meth:`onetick.py.Source.modify_query_times`
    | :py:meth:`onetick.py.Source.time_interval_shift`

    Examples
    --------

    >>> start = otp.dt(2022, 3, 2)
    >>> end = otp.dt(2022, 3, 2) + otp.Milli(3)
    >>> data = otp.DataSource('US_COMP', symbols='AAPL', tick_type='TRD')

    By default, ``time_interval_change()`` does nothing:

    >>> t = data.time_interval_change()
    >>> otp.run(t, start=start, end=end)
                         Time  PRICE  SIZE
    0 2022-03-02 00:00:00.000    1.0   100
    1 2022-03-02 00:00:00.001    1.1   101
    2 2022-03-02 00:00:00.002    1.2   102

    Decreasing time range will not change ticks' timestamps:

    >>> t = data.time_interval_change(start_change=otp.Milli(1), end_change=-otp.Milli(1))
    >>> otp.run(t, start=start, end=end)
                         Time  PRICE  SIZE
    0 2022-03-02 00:00:00.001    1.1   101

    Increasing time range will change timestamps of the ticks that crossed the border.
    In this case first tick's timestamp will be set to original start time,
    and third tick's to original end time.

    >>> t = data.time_interval_change(start_change=-otp.Milli(1), end_change=otp.Milli(1))
    >>> otp.run(t, start=start + otp.Milli(1), end=start + otp.Milli(2))
                         Time  PRICE  SIZE
    0 2022-03-02 00:00:00.001    1.0   100
    1 2022-03-02 00:00:00.001    1.1   101
    2 2022-03-02 00:00:00.002    1.2   102

    Here we are querying empty time interval, but changing start time one second back to get ticks.

    >>> t = data.time_interval_change(start_change=-otp.Second(1))
    >>> otp.run(t, start=start + otp.Second(1), end=end + otp.Second(1))
                     Time  PRICE  SIZE
    0 2022-03-02 00:00:01    1.0   100
    1 2022-03-02 00:00:01    1.1   101
    2 2022-03-02 00:00:01    1.2   102
    """
    start = self['_START_TIME'] + start_change
    end = self['_END_TIME'] + end_change

    # change ticks' timestamps only if they are out of bounds
    output_timestamp = self['TIMESTAMP']
    output_timestamp = otp.math.min(output_timestamp, self['_END_TIME'])
    output_timestamp = otp.math.max(output_timestamp, self['_START_TIME'])

    return self.modify_query_times(start=start, end=end, output_timestamp=output_timestamp, inplace=inplace)
